""" Implements various models.

:author: Michel Bierlaire
:date: Fri Mar 29 17:13:14 2019
"""

# Too constraining
# pylint: disable=invalid-name
# pylint: disable=too-many-lines

import logging
import biogeme.distributions as dist
import biogeme.exceptions as excep
import biogeme.expressions as expr

logger = logging.getLogger(__name__)


def loglogit(V, av, i):
    """The logarithm of the logit model

    The model is defined as

    .. math:: \\frac{a_i e^{V_i}}{\\sum_{i=1}^J a_j e^{V_j}}

    :param V: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.

    :type V: dict(int:biogeme.expressions.expr.Expression)

    :param av: dict of objects representing the availability of each
               alternative (:math:`a_i` in the above formula), indexed
               by numerical ids. Must be consistent with V, or
               None. In this case, all alternatives are supposed to be
               always available.

    :type av: dict(int:biogeme.expressions.expr.Expression)

    :param i: id of the alternative for which the probability must be
              calculated.
    :type i: int

    :return: choice probability of alternative number i.
    :rtype: biogeme.expressions.expr.Expression
    """

    if av is None:
        return expr._bioLogLogitFullChoiceSet(V, av=None, choice=i)

    return expr._bioLogLogit(V, av, i)


def loglogit_sampling(V, av, correction, i):
    """The logarithm of the logit model, with samples of alternatives

    The model is defined as

    .. math:: \\frac{a_i e^{V_i + \\log \\pi_i}}{\\sum_{i=1}^J a_j e^{V_j+\\log \\pi_j}}

    :param V: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.

    :type V: dict(int:biogeme.expressions.expr.Expression)

    :param av: dict of objects representing the availability of each
               alternative (:math:`a_i` in the above formula), indexed
               by numerical ids. Must be consistent with V, or
               None. In this case, all alternatives are supposed to be
               always available.

    :type av: dict(int:biogeme.expressions.expr.Expression)

    :param correction: a dict of expressions for the correstion terms
                       of each alternative.
    :type correction: dict(int:biogeme.expressions.expr.Expression)

    :param i: id of the alternative for which the probability must be
              calculated.
    :type i: int

    :return: choice probability of alternative number i.
    :rtype: biogeme.expressions.expr.Expression
    """
    if V.keys() != correction.keys():
        error_msg = (
            f'The keys of the correction must be the same as the keys of '
            f'the utilities. Correction: {correction.keys()}. '
            f'Utilities: {V.keys}'
        )
        raise excep.BiogemeError(error_msg)
    corrected_V = {k: v - correction[k] for k, v in V.items()}
    return loglogit(corrected_V, av, i)


def logit(V, av, i):
    """The logit model

    The model is defined as

    .. math:: \\frac{a_i e^{V_i}}{\\sum_{i=1}^J a_j e^{V_j}}

    :param V: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.

    :type V: dict(int:biogeme.expressions.expr.Expression)

    :param av: dict of objects representing the availability of each
               alternative (:math:`a_i` in the above formula), indexed
               by numerical ids. Must be consistent with V, or
               None. In this case, all alternatives are supposed to be
               always available.

    :type av: dict(int:biogeme.expressions.expr.Expression)

    :param i: id of the alternative for which the probability must be
              calculated.
    :type i: int

    :return: choice probability of alternative number i.
    :rtype: biogeme.expressions.expr.Expression

    """
    if av is None:
        return expr.exp(expr._bioLogLogitFullChoiceSet(V, av=None, choice=i))

    return expr.exp(expr._bioLogLogit(V, av, i))


def boxcox(x, ell):
    """Box-Cox transform

    .. math:: B(x, \\ell) = \\frac{x^{\\ell}-1}{\\ell}.

    It has the property that

    .. math:: \\lim_{\\ell \\to 0} B(x,\\ell)=\\log(x).

    To avoid numerical difficulties, if :math:`\\ell < 10^{-5}`,
    the McLaurin approximation is used:

    .. math:: \\log(x) + \\ell \\log(x)^2 + \\frac{1}{6} \\ell^2 \\log(x)^3
              + \\frac{1}{24} \\ell^3 \\log(x)^4.

    :param x: a variable to transform.
    :type x: biogeme.expressions.expr.Expression
    :param ell: parameter of the transformation.
    :type ell: biogeme.expressions.expr.Expression

    :return: the Box-Cox transform
    :rtype: biogeme.expressions.expr.Expression
    """
    if isinstance(ell, expr.Beta) and (ell.ub is None or ell.lb is None):
        warning_msg = (
            f'It is advised to set the bounds on parameter {ell.name}. '
            f'A value of -10 and 10 should be appropriate: Beta("{ell.name}", '
            f'{ell.initValue}, -10, 10, {ell.status})'
        )
        logger.warning(warning_msg)

    regular = (x**ell - 1.0) / ell
    mclaurin = (
        expr.log(x)
        + ell * expr.log(x) ** 2
        + ell**2 * expr.log(x) ** 3 / 6.0
        + ell**3 * expr.log(x) ** 4 / 24.0
    )
    close_to_zero = (ell < expr.Numeric(1.0e-5)) * (ell > -expr.Numeric(1.0e-5))
    smooth = expr.Elem({0: regular, 1: mclaurin}, close_to_zero)
    return expr.Elem({0: smooth, 1: expr.Numeric(0)}, x == 0)


def piecewiseVariables(variable, thresholds):
    """Generate the variables to include in a piecewise linear specification.

    If there are K thresholds, K-1 variables are generated. The first
    and last thresholds can be defined as None, corresponding to
    :math:`-\\infty` and :math:`+\\infty`,respectively. If :math:`t` is
    the variable of interest, for each interval :math:`[a:a+b[`, we
    define a variable defined as:

    .. math:: x_{Ti} =\\left\\{  \\begin{array}{ll} 0 & \\text{if }
              t < a \\\\ t-a & \\text{if } a \\leq t < a+b \\\\ b  &
              \\text{otherwise}  \\end{array}\\right. \\;\\;\\;x_{Ti} =
              \\max(0, \\min(t-a, b))

    :param variable: variable for which we need the piecewise linear
       transform. The expression itself or the name of the variable
       can be given.
    :type variable: biogeme.expressions.expr.Expression or str

    :param thresholds: list of thresholds
    :type thresholds: list(float)

    :return: list of variables to for the piecewise linear specification.
    :rtype: list(biogeme.expressions.expr.Expression)

    :raise BiogemeError: if the thresholds are not defined properly,
        as only the first and the last thresholds can be set
        to None.

    .. seealso:: :meth:`piecewiseFormula`

    """
    eye = len(thresholds)
    if all(t is None for t in thresholds):
        errorMsg = (
            'All thresholds for the piecewise linear specification are set to None.'
        )
        raise excep.BiogemeError(errorMsg)
    if None in thresholds[1:-1]:
        errorMsg = (
            'For piecewise linear specification, only the first and '
            'the last thresholds can be None'
        )
        raise excep.BiogemeError(errorMsg)

    # If the name of the variable is given, we transform it into an expression.
    if isinstance(variable, str):
        variable = expr.Variable(variable)

    # First variable
    if thresholds[0] is None:
        results = [expr.bioMin(variable, thresholds[1])]
    else:
        b = thresholds[1] - thresholds[0]
        results = [
            expr.bioMax(expr.Numeric(0), expr.bioMin(variable - thresholds[0], b))
        ]

    for i in range(1, eye - 2):
        b = thresholds[i + 1] - thresholds[i]
        results += [
            expr.bioMax(expr.Numeric(0), expr.bioMin(variable - thresholds[i], b))
        ]

    # Last variable
    if thresholds[-1] is None:
        results += [expr.bioMax(0, variable - thresholds[-2])]
    else:
        b = thresholds[-1] - thresholds[-2]
        results += [
            expr.bioMax(expr.Numeric(0), expr.bioMin(variable - thresholds[-2], b))
        ]
    return results


def piecewiseFormula(variable, thresholds, betas=None):
    """Generate the formula for a piecewise linear specification.

    If there are K thresholds, K-1 variables are generated. The first
    and last thresholds can be defined as None, corresponding to
    :math:`-\\infty` and :math:`+\\infty`, respectively. If :math:`t` is
    the variable of interest, for each interval :math:`[a:a+b[`, we
    define a variable defined as:

    .. math:: x_{Ti} =\\left\\{  \\begin{array}{ll} 0 & \\text{if }
              t < a \\\\ t-a & \\text{if } a \\leq t < a+b \\\\ b  &
              \\text{otherwise}  \\end{array}\\right. \\;\\;\\;x_{Ti} =
              \\max(0, \\min(t-a, b))

    New variables and new parameters are automatically created to
    obtain the specification

    .. math:: \\sum_{i=1}^{K-1} \\beta_i x_{Ti}

    :param variable: name of the variable for which we need the
        piecewise linear transform.
    :type variable: string

    :param thresholds: list of thresholds
    :type thresholds: list(float)

    :param betas: list of beta parameters to be used in the
        specification.  The number of entries should be the number of
        thresholds, minus one. If None, for each interval, the
        parameter Beta('beta_VAR_interval',0, None, None, 0) is used,
        where var is the name of the variable. Default: none.
    :type betas:
        list(biogeme.expresssions.Beta)

    :return: expression of  the piecewise linear specification.
    :rtype: biogeme.expressions.expr.Expression

    :raise BiogemeError: if the thresholds are not defined properly,
        which means that only the first and the last threshold can be set
        to None.

    :raise BiogemeError: if the length of list ``initialexpr.Betas`` is
        not equal to the length of ``thresholds`` minus one.

    .. seealso:: :meth:`piecewiseVariables`

    """
    if isinstance(variable, expr.Variable):
        the_variable = variable
        the_name = variable.name
    elif isinstance(variable, str):
        the_name = variable
        the_variable = expr.Variable(f'{variable}')
    else:
        errorMsg = (
            'The first argument of piecewiseFormula must be the '
            'name of a variable, or the variable itself..'
        )
        raise excep.BiogemeError(errorMsg)

    eye = len(thresholds)
    if all(t is None for t in thresholds):
        errorMsg = (
            'All thresholds for the piecewise linear specification are set to None.'
        )
        raise excep.BiogemeError(errorMsg)
    if None in thresholds[1:-1]:
        errorMsg = (
            'For piecewise linear specification, only the first and '
            'the last thresholds can be None'
        )
        raise excep.BiogemeError(errorMsg)
    if betas is not None:
        if len(betas) != eye - 1:
            errorMsg = (
                f'As there are {eye} thresholds, a total of {eye-1} '
                f'Beta parameters are needed, and not {len(betas)}.'
            )
            raise excep.BiogemeError(errorMsg)

    theVars = piecewiseVariables(the_variable, thresholds)
    if betas is None:
        betas = []
        for i, a_threshold in enumerate(thresholds[:-1]):
            next_threshold = thresholds[i + 1]
            a_name = 'minus_inf' if a_threshold is None else f'{a_threshold}'
            next_name = 'inf' if next_threshold is None else f'{next_threshold}'
            betas.append(
                expr.Beta(f'beta_{the_name}_{a_name}_{next_name}', 0, None, None, 0)
            )

    terms = [beta * theVars[i] for i, beta in enumerate(betas)]

    return expr.bioMultSum(terms)


def piecewise_as_variable(variable, thresholds, betas=None):
    """Generate the formula for a piecewise linear specification, seen
    as a transformed variable.

    If there are K thresholds, K-1 variables are generated. The first
    and last thresholds can be defined as None, corresponding to
    :math:`-\\infty` and :math:`+\\infty`, respectively. If :math:`t` is
    the variable of interest, for each interval :math:`[a:a+b[`, we
    define a variable defined as:

    .. math:: x_{Ti} =\\left\\{  \\begin{array}{ll} 0 & \\text{if }
              t < a \\\\ t-a & \\text{if } a \\leq t < a+b \\\\ b  &
              \\text{otherwise}  \\end{array}\\right. \\;\\;\\;x_{Ti} =
              \\max(0, \\min(t-a, b))

    The specification this is returned is

    .. math:: x_{T1} + \\sum_{i=2}^{K-1} \beta_i x_{Ti}

    :param variable: name of the variable for which we need the
        piecewise linear transform.
    :type variable: string

    :param thresholds: list of thresholds
    :type thresholds: list(float)

    :param betas: list of beta parameters to be used in the
        specification.  The number of entries should be the number of
        thresholds, minus two. If None, for each interval, the
        parameter Beta('beta_VAR_interval',0, None, None, 0) is used,
        where var is the name of the variable. Default: none.
    :type betas:
        list(biogeme.expresssions.Beta)

    :return: expression of  the piecewise linear specification.
    :rtype: biogeme.expressions.expr.Expression

    :raise BiogemeError: if the thresholds are not defined properly,
        which means that only the first and the last threshold can be set
        to None.

    :raise BiogemeError: if the length of list ``initialexpr.Betas`` is
        not equal to the length of ``thresholds`` minus one.

    .. seealso:: :meth:`piecewiseVariables`

    """
    if isinstance(variable, expr.Variable):
        the_variable = variable
        the_name = variable.name
    elif isinstance(variable, str):
        the_name = variable
        the_variable = expr.Variable(f'{variable}')
    else:
        errorMsg = (
            'The first argument of piecewiseFormula must be the '
            'name of a variable, or the variable itself..'
        )
        raise excep.BiogemeError(errorMsg)

    eye = len(thresholds)
    if all(t is None for t in thresholds):
        errorMsg = (
            'All thresholds for the piecewise linear specification are set to None.'
        )
        raise excep.BiogemeError(errorMsg)
    if None in thresholds[1:-1]:
        errorMsg = (
            'For piecewise linear specification, only the first and '
            'the last thresholds can be None'
        )
        raise excep.BiogemeError(errorMsg)
    if betas is not None:
        if len(betas) != eye - 2:
            errorMsg = (
                f'As there are {eye} thresholds, a total of {eye-2} '
                f'Beta parameters are needed, and not {len(betas)}.'
            )
            raise excep.BiogemeError(errorMsg)

    theVars = piecewiseVariables(the_variable, thresholds)
    if betas is None:
        betas = []
        for i, a_threshold in enumerate(thresholds[1:-1]):
            next_threshold = thresholds[i + 2]
            a_name = 'minus_inf' if a_threshold is None else f'{a_threshold}'
            next_name = 'inf' if next_threshold is None else f'{next_threshold}'
            betas.append(
                expr.Beta(f'beta_{the_name}_{a_name}_{next_name}', 0, None, None, 0)
            )

    terms = [beta * theVars[i] for i, beta in enumerate(betas)]

    return theVars[0] + expr.bioMultSum(terms)


def piecewiseFunction(x, thresholds, betas):
    """Plot a piecewise linear specification.

    If there are K thresholds, K-1 variables are generated. The first
    and last thresholds can be defined as None, corresponding to
    :math:`-\\infty` and :math:`+\\infty`, respectively. If :math:`t` is
    the variable of interest, for each interval :math:`[a:a+b[`, we
    define a variable defined as:

    .. math:: x_{Ti} =\\left\\{  \\begin{array}{ll} 0 & \\text{if }
              t < a \\\\ t-a & \\text{if } a \\leq t < a+b \\\\ b  &
              \\text{otherwise}  \\end{array}\\right. \\;\\;\\;x_{Ti} =
              \\max(0, \\min(t-a, b))

    :param x: value at which the piecewise specification must be avaluated
    :type x: float

    :param thresholds: list of thresholds
    :type thresholds: list(float)

    :param betas: list of the beta parameters.  The number of entries
                         should be the number of thresholds, plus
                         one.
    :type betas: list(float)

    :return: value of the numpy function
    :rtype: float

    :raise BiogemeError: if the thresholds are not defined properly,
        which means that only the first and the last threshold can be set
        to None.


    """
    eye = len(thresholds)
    if all(t is None for t in thresholds):
        errorMsg = (
            'All thresholds for the piecewise linear specification ' 'are set to None.'
        )
        raise excep.BiogemeError(errorMsg)
    if None in thresholds[1:-1]:
        errorMsg = (
            'For piecewise linear specification, only the first and '
            'the last thresholds can be None'
        )
        raise excep.BiogemeError(errorMsg)
    if len(betas) != eye - 1:
        errorMsg = (
            f'As there are {eye} thresholds, a total of {eye-1} values '
            f'are needed to initialize the parameters. But '
            f'{len(betas)} are provided'
        )
        raise excep.BiogemeError(errorMsg)

    # If the first threshold is not -infinity, we need to check if
    # x is beyond it.
    if thresholds[0] is not None:
        if x < thresholds[0]:
            return 0
    rest = x
    total = 0
    for i, v in enumerate(betas):
        if thresholds[i + 1] is None:
            total += v * rest
            return total
        if x < thresholds[i + 1]:
            total += v * rest
            return total
        total += v * (
            thresholds[i + 1] - (0 if thresholds[i] is None else thresholds[i])
        )
        rest = x - thresholds[i + 1]
    return total


def logmev(V, logGi, av, choice):
    """Log of the choice probability for a MEV model.

    :param V: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.

    :type V: dict(int:biogeme.expressions.expr.Expression)

    :param logGi: a dictionary mapping each alternative id with the function

        .. math:: \\ln \\frac{\\partial G}{\\partial y_i}
            (e^{V_1},\\ldots,e^{V_J})

        where :math:`G` is the MEV generating function. If an alternative
        :math:`i` is not available, then :math:`G_i = 0`.

    :type logGi: dict(int:biogeme.expressions.expr.Expression)

    :param av: dict of objects representing the availability of each
               alternative (:math:`a_i` in the above formula), indexed
               by numerical ids. Must be consistent with V, or
               None. In this case, all alternatives are supposed to be
               always available.

    :type av: dict(int:biogeme.expressions.expr.Expression)

    :param choice: id of the alternative for which the probability must be
              calculated.
    :type choice: biogeme.expressions.expr.Expression

    :return: log of the choice probability of the MEV model, given by
    :rtype: biogeme.expressions.expr.Expression

    .. math:: V_i + \\ln G_i(e^{V_1},\\ldots,e^{V_J}) -
              \\ln\\left(\\sum_j e^{V_j + \\ln G_j(e^{V_1},
              \\ldots,e^{V_J})}\\right)

    """
    H = {i: v + logGi[i] for i, v in V.items()}
    if av is None:
        logP = expr._bioLogLogitFullChoiceSet(H, av=None, choice=choice)
    else:
        logP = expr._bioLogLogit(H, av, choice)
    return logP


def mev(V, logGi, av, choice):
    """Choice probability for a MEV model.

    :param V: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.

    :type V: dict(int:biogeme.expressions.expr.Expression)


    :param logGi: a dictionary mapping each alternative id with the function

        .. math:: \\ln \\frac{\\partial G}{\\partial y_i}
              (e^{V_1}, \\ldots, e^{V_J})

        where :math:`G` is the MEV generating function. If an alternative
        :math:`i` is not available, then :math:`G_i = 0`.

    :type logGi: dict(int:biogeme.expressions.expr.Expression)

    :param av: dict of objects representing the availability of each
               alternative (:math:`a_i` in the above formula), indexed
               by numerical ids. Must be consistent with V, or
               None. In this case, all alternatives are supposed to be
               always available.

    :type av: dict(int:biogeme.expressions.expr.Expression)

    :param choice: id of the alternative for which the probability must be
              calculated.
    :type choice: biogeme.expressions.expr.Expression

    :return: Choice probability of the MEV model, given by

    .. math:: \\frac{e^{V_i + \\ln G_i(e^{V_1},
              \\ldots,e^{V_J})}}{\\sum_j e^{V_j +
              \\ln G_j(e^{V_1},\\ldots,e^{V_J})}}

    :rtype: biogeme.expressions.expr.Expression
    """
    return expr.exp(logmev(V, logGi, av, choice))


def logmev_endogenousSampling(V, logGi, av, correction, choice):
    """Log of choice probability for a MEV model, including the
    correction for endogenous sampling as proposed by `Bierlaire, Bolduc
    and McFadden (2008)`_.

    .. _`Bierlaire, Bolduc and McFadden (2008)`:
       http://dx.doi.org/10.1016/j.trb.2007.09.003

    :param V: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.

    :type V: dict(int:biogeme.expressions.expr.Expression)

    :param logGi: a dictionary mapping each alternative id with the function

        .. math:: \\ln \\frac{\\partial G}{\\partial y_i}
                  (e^{V_1}, \\ldots, e^{V_J})

        where :math:`G` is the MEV generating function. If an alternative
        :math:`i` is not available, then :math:`G_i = 0`.

    :type logGi: dict(int:biogeme.expressions.expr.Expression)

    :param av: dict of objects representing the availability of each
               alternative (:math:`a_i` in the above formula), indexed
               by numerical ids. Must be consistent with V, or
               None. In this case, all alternatives are supposed to be
               always available.

    :type av: dict(int:biogeme.expressions.expr.Expression)


    :param correction: a dict of expressions for the correstion terms
                       of each alternative.
    :type correction: dict(int:biogeme.expressions.expr.Expression)

    :param choice: id of the alternative for which the probability must be
              calculated.
    :type choice: biogeme.expressions.expr.Expression

    :return: log of the choice probability of the MEV model, given by

    .. math:: V_i + \\ln G_i(e^{V_1}, \\ldots,e^{V_J}) + \\omega_i -
              \\ln\\left(\\sum_j e^{V_j +
              \\ln G_j(e^{V_1}, \\ldots, e^{V_J})+ \\omega_j}\\right)

    where :math:`\\omega_i` is the correction term for alternative :math:`i`.

    :rtype: biogeme.expressions.expr.Expression

    """
    H = {i: v + logGi[i] + correction[i] for i, v in V.items()}
    logP = expr._bioLogLogit(H, av, choice)
    return logP


def mev_endogenousSampling(V, logGi, av, correction, choice):
    """Choice probability for a MEV model, including the correction
    for endogenous sampling as proposed by
    `Bierlaire, Bolduc and McFadden (2008)`_.

    .. _`Bierlaire, Bolduc and McFadden (2008)`:
           http://dx.doi.org/10.1016/j.trb.2007.09.003

    :param V: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.

    :type V: dict(int:biogeme.expressions.expr.Expression)

    :param logGi: a dictionary mapping each alternative id with the function

        .. math:: \\ln \\frac{\\partial G}{\\partial y_i}
              (e^{V_1}, \\ldots, e^{V_J})

        where :math:`G` is the MEV generating function. If an alternative
        :math:`i` is not available, then :math:`G_i = 0`.

    :type logGi: dict(int:biogeme.expressions.expr.Expression)

    :param av: dict of objects representing the availability of each
               alternative (:math:`a_i` in the above formula), indexed
               by numerical ids. Must be consistent with V, or
               None. In this case, all alternatives are supposed to be
               always available.

    :type av: dict(int:biogeme.expressions.expr.Expression)


    :param correction: a dict of expressions for the correstion terms
                       of each alternative.
    :type correction: dict(int:biogeme.expressions.expr.Expression)

    :param choice: id of the alternative for which the probability must be
              calculated.
    :type choice: biogeme.expressions.expr.Expression

    :return: log of the choice probability of the MEV model, given by

    .. math:: V_i + \\ln G_i(e^{V_1}, \\ldots, e^{V_J}) + \\omega_i -
              \\ln\\left(\\sum_j e^{V_j + \\ln G_j(e^{V_1},\\ldots,e^{V_J})+
              \\omega_j}\\right)

    where :math:`\\omega_i` is the correction term for alternative :math:`i`.

    :rtype: biogeme.expressions.expr.Expression

    """
    return expr.exp(logmev_endogenousSampling(V, logGi, av, correction, choice))


def getMevGeneratingForNested(V, availability, nests, alone=None):
    """Implements the  MEV generating function for the nested logit model

    :param V: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.
    :type V: dict(int:biogeme.expressions.expr.Expression)

    :param availability: dict of objects representing the availability of each
               alternative, indexed
               by numerical ids. Must be consistent with V, or
               None. In this case, all alternatives are supposed to be
               always available.

    :type availability: dict(int:biogeme.expressions.expr.Expression)

    :param nests: A tuple containing as many items as nests.
        Each item is also a tuple containing two items:

        - an object of type biogeme.expressions.expr.Expression representing
          the nest parameter,
        - a list containing the list of identifiers of the alternatives
          belonging to the nest.

        Example::

            nesta = MUA ,[1, 2, 3]
            nestb = MUB ,[4, 5, 6]
            nests = nesta, nestb

    :type nests: tuple

    :param alone: list of alternatives that are alone in a nest
    :type alone: list(int)

    :return: a dictionary mapping each alternative id with the function

    .. math:: G(e^{V_1},
              \\ldots,e^{V_J}) =  \\sum_m \\left( \\sum_{\\ell \\in C_m}
              y_\\ell^{\\mu_m}\\right)^{\\frac{\\mu}{\\mu_m}}

    where :math:`G` is the MEV generating function.

    :rtype: biogeme.expressions.expr.Expression

    """

    termsForNests = []
    for m in nests:
        if availability is None:
            sumdict = [expr.exp(m[0] * V[i]) for i in m[1]]
        else:
            sumdict = [
                expr.Elem(
                    {0: 0.0, 1: expr.exp(m[0] * V[i])},
                    availability[i] != expr.Numeric(0),
                )
                for i in m[1]
            ]
        theSum = expr.bioMultSum(sumdict)
        termsForNests.append(theSum**1.0 / m[0])
    if alone is not None:
        for i in alone:
            termsForNests.append(V[i])
    return expr.bioMultSum(termsForNests)


def getMevForNested(V, availability, nests, alone=None):
    """Implements the derivatives of MEV generating function for the
    nested logit model

    :param V: dict of objects representing the utility functions of
        each alternative, indexed by numerical ids.
    :type V: dict(int:biogeme.expressions.expr.Expression)

    :param availability: dict of objects representing the availability of each
               alternative, indexed
               by numerical ids. Must be consistent with V, or
               None. In this case, all alternatives are supposed to be
               always available.

    :type availability: dict(int:biogeme.expressions.expr.Expression)

    :param nests: A tuple containing as many items as nests.
        Each item is also a tuple containing two items:

        - an object of type biogeme.expressions.expr.Expression representing
          the nest parameter,
        - a list containing the list of identifiers of the alternatives
          belonging to the nest.

        Example::

            nesta = MUA ,[1, 2, 3]
            nestb = MUB ,[4, 5, 6]
            nests = nesta, nestb

    :type nests: tuple

    :param alone: list of alternatives that are alone in a nest
    :type alone: list(int)

    :return: a dictionary mapping each alternative id with the function

        .. math:: \\ln \\frac{\\partial G}{\\partial y_i}(e^{V_1},
              \\ldots,e^{V_J}) = e^{(\\mu_m-1)V_i}
              \\left(\\sum_{i=1}^{J_m} e^{\\mu_m V_i}\\right)^
              {\\frac{1}{\\mu_m}-1}

        where :math:`m` is the (only) nest containing alternative :math:`i`,
        and :math:`G` is the MEV generating function.

    :rtype: dict(int:biogeme.expressions.expr.Expression)

    """

    if alone is None:
        logGi = {}
    else:
        logGi = {i: 0 for i in alone}
    for m in nests:
        if availability is None:
            sumdict = [expr.exp(m[0] * V[i]) for i in m[1]]
        else:
            sumdict = [
                expr.Elem(
                    {0: 0.0, 1: expr.exp(m[0] * V[i])},
                    availability[i] != expr.Numeric(0),
                )
                for i in m[1]
            ]
        theSum = expr.bioMultSum(sumdict)
        for i in m[1]:
            logGi[i] = (m[0] - 1.0) * V[i] + (1.0 / m[0] - 1.0) * expr.log(theSum)
    return logGi


def getMevForNestedMu(V, availability, nests, mu, alone=None):
    """Implements the MEV generating function for the nested logit model,
    including the scale parameter

    :param V: dict of objects representing the utility functions of
        each alternative, indexed by numerical ids.

    :type V: dict(int:biogeme.expressions.expr.Expression)

    :param availability: dict of objects representing the availability
        of each alternative, indexed
        by numerical ids. Must be consistent with V, or
        None. In this case, all alternatives are supposed to be
        always available.

    :type availability: dict(int:biogeme.expressions.expr.Expression)

    :param nests: A tuple containing as many items as nests.
        Each item is also a tuple containing two items:

        - an object of type biogeme.expressions. expr.Expression
          representing the nest parameter,
        - a list containing the list of identifiers of the alternatives
          belonging to the nest.

        Example::

            nesta = MUA, [1, 2, 3]
           nestb = MUB, [4, 5, 6]
            nests = nesta, nestb

    :type nests: tuple

    :param mu: scale parameter
    :type mu: biogeme.expressions.expr.Expression

    :param alone: list of alternatives that are alone in a nest
    :type alone: list(int)

    :return: a dictionary mapping each alternative id with the function

        .. math:: \\frac{\\partial G}{\\partial y_i}(e^{V_1},\\ldots,e^{V_J}) =
                  \\mu e^{(\\mu_m-1)V_i} \\left(\\sum_{i=1}^{J_m}
                  e^{\\mu_m V_i}\\right)^{\\frac{\\mu}{\\mu_m}-1}

        where :math:`m` is the (only) nest containing alternative :math:`i`,
        and :math:`G` is the MEV generating function.

    :rtype: dict(int:biogeme.expressions.expr.Expression)

    """

    if alone is None:
        logGi = {}
    else:
        logGi = {i: expr.log(mu) + (mu - 1) * V[i] for i in alone}
    for m in nests:
        if availability is None:
            sumdict = [expr.exp(m[0] * V[i]) for i in m[1]]
        else:
            sumdict = [
                expr.Elem({0: 0.0, 1: expr.exp(m[0] * V[i])}, availability[i] != 0)
                for i in m[1]
            ]
        theSum = expr.bioMultSum(sumdict)
        for i in m[1]:
            logGi[i] = (
                expr.log(mu)
                + (m[0] - 1.0) * V[i]
                + (mu / m[0] - 1.0) * expr.log(theSum)
            )
    return logGi


def nested(V, availability, nests, choice, alone=None):
    """Implements the nested logit model as a MEV model.

    :param V: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.
    :type V: dict(int:biogeme.expressions.expr.Expression)

    :param availability: dict of objects representing the availability
                         of each alternative, indexed by numerical
                         ids. Must be consistent with V, or None. In
                         this case, all alternatives are supposed to
                         be always available.

    :type availability: dict(int:biogeme.expressions.expr.Expression)

    :param nests: A tuple containing as many items as nests. Each item is also
        a tuple containing two items:

        - an object of type biogeme.expressions. expr.Expression
          representing the nest parameter,
        - a list containing the list of identifiers of the
          alternatives belonging to the nest.

        Example::

            nesta = MUA, [1, 2, 3]
            nestb = MUB, [4, 5, 6]
            nests = nesta, nestb

    :type nests: tuple

    :param choice: id of the alternative for which the probability must be
              calculated.
    :type choice: biogeme.expressions.expr.Expression

    :param alone: list of alternatives that are alone in a nest
    :type alone: list(int)

    :return: choice probability for the nested logit model,
             based on the derivatives of the MEV generating function produced
             by the function getMevForNested

    :rtype: biogeme.expressions.expr.Expression

    :raise BiogemeError: if the definition of the nests is invalid.
    """

    ok, message = checkValidityNestedLogit(V, nests, alone)
    if not ok:
        raise excep.BiogemeError(message)

    logGi = getMevForNested(V, availability, nests, alone)
    P = mev(V, logGi, availability, choice)
    return P


def lognested(V, availability, nests, choice, alone=None):
    """Implements the log of a nested logit model as a MEV model.

    :param V: dict of objects representing the utility functions of
        each alternative, indexed by numerical ids.

    :type V: dict(int:biogeme.expressions.expr.Expression)

    :param availability: dict of objects representing the availability of each
        alternative (:math:`a_i` in the above formula), indexed
        by numerical ids. Must be consistent with V, or
        None. In this case, all alternatives are supposed to be
        always available.

    :type availability: dict(int:biogeme.expressions.expr.Expression)

    :param nests: A tuple containing as many items as nests.
        Each item is also a tuple containing two items:

        - an object of type biogeme.expressions. expr.Expression representing
          the nest parameter,
        - a list containing the list of identifiers of the alternatives
          belonging to the nest.

        Example::

            nesta = MUA, [1, 2, 3]
            nestb = MUB, [4, 5, 6]
            nests = nesta, nestb

    :type nests: tuple

    :param choice: id of the alternative for which the probability must be
              calculated.
    :type choice: biogeme.expressions.expr.Expression

    :param alone: list of alternatives that are alone in a nest
    :type alone: list(int)

    :return: log of choice probability for the nested logit model,
             based on the derivatives of the MEV generating function produced
             by the function getMevForNested

    :rtype: biogeme.expressions.expr.Expression

    :raise BiogemeError: if the definition of the nests is invalid.
    """
    ok, message = checkValidityNestedLogit(V, nests, alone)
    if not ok:
        raise excep.BiogemeError(message)
    logGi = getMevForNested(V, availability, nests, alone)
    logP = logmev(V, logGi, availability, choice)
    return logP


def nestedMevMu(V, availability, nests, choice, mu, alone=None):
    """Implements the nested logit model as a MEV model, where mu is also
    a parameter, if the user wants to test different normalization
    schemes.

    :param V: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.

    :type V: dict(int:biogeme.expressions.expr.Expression)

    :param availability: dict of objects representing the availability of each
               alternative (:math:`a_i` in the above formula), indexed
               by numerical ids. Must be consistent with V, or
               None. In this case, all alternatives are supposed to be
               always available.

    :type availability: dict(int:biogeme.expressions.expr.Expression)

    :param nests: A tuple containing as many items as nests.
        Each item is also a tuple containing two items:

        - an object of type biogeme.expressions.expr.Expression  representing
          the nest parameter,
        - a list containing the list of identifiers of the alternatives
          belonging to the nest.

        Example::

            nesta = MUA ,[1, 2, 3]
            nestb = MUB ,[4, 5, 6]
            nests = nesta, nestb

    :type nests: tuple

    :param choice: id of the alternative for which the probability must be
              calculated.
    :type choice: biogeme.expressions.expr.Expression

    :param mu: expression producing the value of the top-level scale parameter.
    :type mu:  biogeme.expressions.expr.Expression

    :param alone: list of alternatives that are alone in a nest
    :type alone: list(int)

    :return: the nested logit choice probability based on the following
             derivatives of the MEV generating function:

    .. math:: \\frac{\\partial G}{\\partial y_i}(e^{V_1},\\ldots,e^{V_J}) =
              \\mu e^{(\\mu_m-1)V_i} \\left(\\sum_{i=1}^{J_m}
              e^{\\mu_m V_i}\\right)^{\\frac{\\mu}{\\mu_m}-1}

    Where :math:`m` is the (only) nest containing alternative :math:`i`, and
    :math:`G` is the MEV generating function.

    :rtype: biogeme.expressions.expr.Expression

    """
    return expr.exp(lognestedMevMu(V, availability, nests, choice, mu, alone))


def lognestedMevMu(V, availability, nests, choice, mu, alone=None):
    """Implements the log of the nested logit model as a MEV model, where
    mu is also a parameter, if the user wants to test different
    normalization schemes.


    :param V: dict of objects representing the utility functions of
        each alternative, indexed by numerical ids.

    :type V: dict(int:biogeme.expressions.expr.Expression)

    :param availability: dict of objects representing the availability of each
               alternative (:math:`a_i` in the above formula), indexed
               by numerical ids. Must be consistent with V, or
               None. In this case, all alternatives are supposed to be
               always available.

    :type availability: dict(int:biogeme.expressions.expr.Expression)

    :param nests: A tuple containing as many items as nests.
        Each item is also a tuple containing two items:

        - an object of type biogeme.expressions.expr.Expression  representing
          the nest parameter,
        - a list containing the list of identifiers of the alternatives
          belonging to the nest.

        Example::

            nesta = MUA, [1, 2, 3]
            nestb = MUB, [4, 5, 6]
            nests = nesta, nestb

    :type nests: tuple

    :param choice: id of the alternative for which the probability must be
              calculated.
    :type choice: biogeme.expressions.expr.Expression

    :param mu: expression producing the value of the top-level scale parameter.
    :type mu:  biogeme.expressions.expr.Expression

    :param alone: list of alternatives that are alone in a nest
    :type alone: list(int)

    :return: the log of the nested logit choice probability based on the
        following derivatives of the MEV generating function:

        .. math:: \\frac{\\partial G}{\\partial y_i}(e^{V_1},\\ldots,e^{V_J}) =
                  \\mu e^{(\\mu_m-1)V_i} \\left(\\sum_{i=1}^{J_m}
                  e^{\\mu_m V_i}\\right)^{\\frac{\\mu}{\\mu_m}-1}

        where :math:`m` is the (only) nest containing alternative :math:`i`,
        and :math:`G` is the MEV generating function.

    :rtype: biogeme.expressions.expr.Expression

    """

    logGi = getMevForNestedMu(V, availability, nests, mu, alone)
    logP = logmev(V, logGi, availability, choice)
    return logP


def cnl_avail(
    V, availability, nests, choice, alone=None, sampling_log_probability=None
):
    """Same as cnl. Maintained for backward compatibility

    :param V: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.
    :type V: dict(int:biogeme.expressions.expr.Expression)

    :param availability: dict of objects representing the availability of each
               alternative, indexed
               by numerical ids. Must be consistent with V, or
               None. In this case, all alternatives are supposed to be
               always available.

    :type availability: dict(int:biogeme.expressions.expr.Expression)

    :param nests: a tuple containing as many items as nests. Each item is
        also a tuple containing two items

        - an object of type biogeme.expressions.expr.Expression  representing
          the nest parameter,
        - a dictionary mapping the alternative ids with the cross-nested
          parameters for the corresponding nest. If an alternative is
          missing in the dictionary, the corresponding alpha is set to zero.

        Example::

            alphaA = {1: alpha1a,
                      2: alpha2a,
                      3: alpha3a,
                      4: alpha4a,
                      5: alpha5a,
                      6: alpha6a}
            alphaB = {1: alpha1b,
                      2: alpha2b,
                      3: alpha3b,
                      4: alpha4b,
                      5: alpha5b,
                      6: alpha6b}
            nesta = MUA, alphaA
            nestb = MUB, alphaB
            nests = nesta, nestb

    :type nests: tuple

    :param choice: id of the alternative for which the probability must be
              calculated.
    :type choice: biogeme.expressions.expr.Expression

    :param alone: list of alternatives that are alone in a nest
    :type alone: list(int)

    :param sampling_log_probability: if not None, it means that the
        choice set is actually a subset that has been sampled from the
        full choice set. In that case, this is a dictionary mapping
        each alternative with the logarithm of its probability to be
        selected in the sample.
    :type sampling_log_probability: dict(int: biogeme.expressions.Expression)

    :return: choice probability for the cross-nested logit model.
    :rtype: biogeme.expressions.expr.Expression
    """
    return cnl(V, availability, nests, choice, alone, sampling_log_probability)


def cnl(V, availability, nests, choice, alone=None, sampling_log_probability=None):
    """Implements the cross-nested logit model as a MEV model.

    :param V: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.
    :type V: dict(int:biogeme.expressions.expr.Expression)

    :param availability: dict of objects representing the availability of each
               alternative, indexed
               by numerical ids. Must be consistent with V, or
               None. In this case, all alternatives are supposed to be
               always available.

    :type availability: dict(int:biogeme.expressions.expr.Expression)

    :param nests: a tuple containing as many items as nests.
        Each item is also a tuple containing two items:

        - an object of type biogeme.expressions. expr.Expression
          representing the nest parameter,
        - a dictionary mapping the alternative ids with the cross-nested
          parameters for the corresponding nest. If an alternative is
          missing in the dictionary, the corresponding alpha is set to zero.

        Example::

            alphaA = {1: alpha1a,
                      2: alpha2a,
                      3: alpha3a,
                      4: alpha4a,
                      5: alpha5a,
                      6: alpha6a}
            alphaB = {1: alpha1b,
                      2: alpha2b,
                      3: alpha3b,
                      4: alpha4b,
                      5: alpha5b,
                      6: alpha6b}
            nesta = MUA, alphaA
            nestb = MUB, alphaB
            nests = nesta, nestb

    :type nests: tuple

    :param sampling_log_probability: if not None, it means that the
        choice set is actually a subset that has been sampled from the
        full choice set. In that case, this is a dictionary mapping
        each alternative with the logarithm of its probability to be
        selected in the sample.
    :type sampling_log_probability: dict(int: biogeme.expressions.Expression)

    :param choice: id of the alternative for which the probability must be
              calculated.
    :type choice: biogeme.expressions.expr.Expression

    :param alone: list of alternatives that are alone in a nest
    :type alone: list(int)

    :return: choice probability for the cross-nested logit model.
    :rtype: biogeme.expressions.expr.Expression
    """
    return expr.exp(
        logcnl(V, availability, nests, choice, alone, sampling_log_probability)
    )


def logcnl_avail(
    V, availability, nests, choice, alone=None, sampling_log_probability=None
):
    """Same as logcnl. Maintained for backward compatibility

    :param V: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.
    :type V: dict(int:biogeme.expressions.expr.Expression)

    :param availability: dict of objects representing the availability of each
               alternative, indexed
               by numerical ids. Must be consistent with V, or
               None. In this case, all alternatives are supposed to be
               always available.

    :type availability: dict(int:biogeme.expressions.expr.Expression)

    :param nests: a tuple containing as many items as nests.
        Each item is also a tuple containing two items:

        - an object of type biogeme.expressions. expr.Expression
          representing the nest parameter,
        - a dictionary mapping the alternative ids with the cross-nested
          parameters for the corresponding nest. If an alternative is
          missing in the dictionary, the corresponding alpha is set to zero.

        Example::

            alphaA = {1: alpha1a,
                      2: alpha2a,
                      3: alpha3a,
                      4: alpha4a,
                      5: alpha5a,
                      6: alpha6a}
            alphaB = {1: alpha1b,
                      2: alpha2b,
                      3: alpha3b,
                      4: alpha4b,
                      5: alpha5b,
                      6: alpha6b}
            nesta = MUA, alphaA
            nestb = MUB, alphaB
            nests = nesta, nestb

    :type nests: tuple

    :param choice: id of the alternative for which the probability must be
              calculated.
    :type choice: biogeme.expressions.expr.Expression

    :param sampling_log_probability: if not None, it means that the
        choice set is actually a subset that has been sampled from the
        full choice set. In that case, this is a dictionary mapping
        each alternative with the logarithm of its probability to be
        selected in the sample.
    :type sampling_log_probability: dict(int: biogeme.expressions.Expression)

    :param alone: list of alternatives that are alone in a nest
    :type alone: list(int)

    :return: log of choice probability for the cross-nested logit model.
    :rtype: biogeme.expressions.expr.Expression
    """
    return logcnl(V, availability, nests, choice, alone, sampling_log_probability)


def getMevForCrossNested(
    V, availability, nests, alone=None, sampling_log_probability=None
):
    """Implements the MEV generating function for the cross nested logit
    model as a MEV model.

    :param V: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.
    :type V: dict(int: biogeme.expressions.expr.Expression)

    :param availability: dict of objects representing the availability of each
        alternative, indexed
        by numerical ids. Must be consistent with V, or
        None. In this case, all alternatives are supposed to be
        always available.

    :type availability: dict(int: biogeme.expressions.Expression)

    :param nests: a tuple containing as many items as nests.
        Each item is also a tuple containing two items:

        - an object of type biogeme.expressions. expr.Expression
          representing the nest parameter,
        - a dictionary mapping the alternative ids with the cross-nested
          parameters for the corresponding nest. If an alternative is
          missing in the dictionary, the corresponding alpha is set to zero.

        Example::

            alphaA = {1: alpha1a,
                      2: alpha2a,
                      3: alpha3a,
                      4: alpha4a,
                      5: alpha5a,
                      6: alpha6a}
            alphaB = {1: alpha1b,
                      2: alpha2b,
                      3: alpha3b,
                      4: alpha4b,
                      5: alpha5b,
                      6: alpha6b}
            nesta = MUA, alphaA
            nestb = MUB, alphaB
            nests = nesta, nestb

    :type nests: tuple

    :param alone: list of alternatives that are alone in a nest
    :type alone: list(int)

    :param sampling_log_probability: if not None, it means that the
        choice set is actually a subset that has been sampled from the
        full choice set. In that case, this is a dictionary mapping
        each alternative with the logarithm of its probability to be
        selected in the sample.
    :type sampling_log_probability: dict(int: biogeme.expressions.Expression)

    :return: log of the choice probability for the cross-nested logit model.
    :rtype: biogeme.expressions.expr.Expression

    """

    Gi_terms = {}
    if alone is None:
        logGi = {}
        for i in V:
            Gi_terms[i] = []
    else:
        logGi = {i: 0 for i in alone}
        for i in set(V).difference(set(alone)):
            Gi_terms[i] = []
    biosum = {}
    for m in nests:
        if availability is None:
            if sampling_log_probability is None:
                biosum = expr.bioMultSum(
                    [a ** (m[0]) * expr.exp(m[0] * (V[i])) for i, a in m[1].items()]
                )
            else:
                biosum = expr.bioMultSum(
                    [
                        expr.exp(-sampling_log_probability[i])
                        * a ** (m[0])
                        * expr.exp(m[0] * (V[i]))
                        for i, a in m[1].items()
                    ]
                )
        else:
            if sampling_log_probability is None:
                biosum = expr.bioMultSum(
                    [
                        availability[i] * a ** (m[0]) * expr.exp(m[0] * (V[i]))
                        for i, a in m[1].items()
                    ]
                )
            else:
                biosum = expr.bioMultSum(
                    [
                        expr.exp(-sampling_log_probability[i])
                        * availability[i]
                        * a ** (m[0])
                        * expr.exp(m[0] * (V[i]))
                        for i, a in m[1].items()
                    ]
                )
        for i, a in m[1].items():
            Gi_terms[i] += [
                a ** (m[0])
                * expr.exp((m[0] - 1) * (V[i]))
                * biosum ** ((1.0 / m[0]) - 1.0)
            ]
    for k, G in Gi_terms.items():
        logGi[k] = expr.Elem(
            {1: 0, 0: expr.log(expr.bioMultSum(G))}, expr.bioMultSum(G) == 0
        )
        if sampling_log_probability is not None:
            logGi[k] -= sampling_log_probability[k]
    return logGi


def logcnl(V, availability, nests, choice, alone=None, sampling_log_probability=None):
    """Implements the log of the cross-nested logit model as a MEV model.

    :param V: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.
    :type V: dict(int:biogeme.expressions.expr.Expression)

    :param availability: dict of objects representing the availability of each
               alternative, indexed
               by numerical ids. Must be consistent with V, or
               None. In this case, all alternatives are supposed to be
               always available.

    :type availability: dict(int:biogeme.expressions.expr.Expression)

    :param nests: a tuple containing as many items as nests.
        Each item is also a tuple containing two items:

        - an object of type biogeme.expressions. expr.Expression
          representing the nest parameter,
        - a dictionary mapping the alternative ids with the cross-nested
          parameters for the corresponding nest. If an alternative is
          missing in the dictionary, the corresponding alpha is set to zero.

        Example::

            alphaA = {1: alpha1a,
                      2: alpha2a,
                      3: alpha3a,
                      4: alpha4a,
                      5: alpha5a,
                      6: alpha6a}
            alphaB = {1: alpha1b,
                      2: alpha2b,
                      3: alpha3b,
                      4: alpha4b,
                      5: alpha5b,
                      6: alpha6b}
            nesta = MUA , alphaA
            nestb = MUB , alphaB
            nests = nesta, nestb

    :type nests: tuple

    :param choice: id of the alternative for which the probability must be
              calculated.
    :type choice: biogeme.expressions.expr.Expression

    :param alone: list of alternatives that are alone in a nest
    :type alone: list(int)

    :param sampling_log_probability: if not None, it means that the
        choice set is actually a subset that has been sampled from the
        full choice set. In that case, this is a dictionary mapping
        each alternative with the logarithm of its probability to be
        selected in the sample.
    :type sampling_log_probability: dict(int: biogeme.expressions.Expression)

    :return: log of the choice probability for the cross-nested logit model.
    :rtype: biogeme.expressions.expr.Expression

    :raise BiogemeError: if the definition of the nests is invalid.
    """
    ok, message = checkValidityCNL(V, nests, alone)
    if not ok:
        raise excep.BiogemeError(message)
    if message != '':
        logger.warning(f'CNL: {message}')
    logGi = getMevForCrossNested(
        V, availability, nests, alone, sampling_log_probability
    )
    logP = logmev(V, logGi, availability, choice)
    return logP


def cnlmu(V, availability, nests, choice, mu, alone=None):
    """Implements the cross-nested logit model as a MEV model with
    the homogeneity parameters is explicitly involved

    :param V: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.
    :type V: dict(int:biogeme.expressions.expr.Expression)

    :param availability: dict of objects representing the availability of each
               alternative, indexed
               by numerical ids. Must be consistent with V, or
               None. In this case, all alternatives are supposed to be
               always available.

    :type availability: dict(int:biogeme.expressions.expr.Expression)

    :param nests: a tuple containing as many items as nests. Each
        item is also a tuple containing two items:

        - an object of type biogeme.expressions. expr.Expression representing
          the nest parameter,
        - a dictionary mapping the alternative ids with the cross-nested
          parameters for the corresponding nest. If an alternative is
          missing in the dictionary, the corresponding alpha is set to zero.

        Example::

            alphaA = {1: alpha1a,
                      2: alpha2a,
                      3: alpha3a,
                      4: alpha4a,
                      5: alpha5a,
                      6: alpha6a}
            alphaB = {1: alpha1b,
                      2: alpha2b,
                      3: alpha3b,
                      4: alpha4b,
                      5: alpha5b,
                      6: alpha6b}
            nesta = MUA, alphaA
            nestb = MUB, alphaB
            nests = nesta, nestb

    :type nests: tuple

    :param choice: id of the alternative for which the probability must be
              calculated.
    :type choice: biogeme.expressions.expr.Expression

    :param mu: Homogeneity parameter :math:`\\mu`.
    :type mu: biogeme.expressions.expr.Expression

    :param alone: list of alternatives that are alone in a nest
    :type alone: list(int)

    :return: choice probability for the cross-nested logit model.
    :rtype: biogeme.expressions.expr.Expression
    """
    return expr.exp(logcnlmu(V, availability, nests, choice, mu, alone))


def getMevForCrossNestedMu(V, availability, nests, mu, alone=None):
    """Implements the MEV generating function for the cross-nested logit
    model as a MEV model with the homogeneity parameters is explicitly
    involved.

    :param V: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.
    :type V: dict(int:biogeme.expressions.expr.Expression)

    :param availability: dict of objects representing the availability of each
               alternative, indexed
               by numerical ids. Must be consistent with V, or
               None. In this case, all alternatives are supposed to be
               always available.

    :type availability: dict(int:biogeme.expressions.expr.Expression)

    :param nests: a tuple containing as many items as nests.
        Each item is also a tuple containing two items:

        - an object of type biogeme.expressions. expr.Expression representing
          the nest parameter,
        - a dictionary mapping the alternative ids with the cross-nested
          parameters for the corresponding nest. If an alternative is
          missing in the dictionary, the corresponding alpha is set to zero.

        Example::

            alphaA = {1: alpha1a,
                      2: alpha2a,
                      3: alpha3a,
                      4: alpha4a,
                      5: alpha5a,
                      6: alpha6a}
            alphaB = {1: alpha1b,
                      2: alpha2b,
                      3: alpha3b,
                      4: alpha4b,
                      5: alpha5b,
                      6: alpha6b}
            nesta = MUA, alphaA
            nestb = MUB, alphaB
            nests = nesta, nestb

    :type nests: tuple

    :param mu: Homogeneity parameter :math:`\\mu`.
    :type mu: biogeme.expressions.expr.Expression

    :param alone: list of alternatives that are alone in a nest
    :type alone: list(int)

    :return: log of the choice probability for the cross-nested logit model.
    :rtype: biogeme.expressions.expr.Expression

    """
    Gi_terms = {}
    if alone is None:
        logGi = {}
        for i in V:
            Gi_terms[i] = []
    else:
        logGi = {i: expr.log(mu) + (mu - 1) * V[i] for i in alone}
        for i in set(V).difference(set(alone)):
            Gi_terms[i] = []
    biosum = {}
    for m in nests:
        if availability is None:
            biosum = expr.bioMultSum(
                [a ** (m[0] / mu) * expr.exp(m[0] * (V[i])) for i, a in m[1].items()]
            )
        else:
            biosum = expr.bioMultSum(
                [
                    availability[i] * a ** (m[0] / mu) * expr.exp(m[0] * (V[i]))
                    for i, a in m[1].items()
                ]
            )
        for i, a in m[1].items():
            Gi_terms[i] += [
                a ** (m[0] / mu)
                * expr.exp((m[0] - 1) * (V[i]))
                * biosum ** ((mu / m[0]) - 1.0)
            ]
    for k, G in Gi_terms.items():
        logGi[k] = expr.log(mu * expr.bioMultSum(G))
    return logGi


def logcnlmu(V, availability, nests, choice, mu, alone=None):
    """Implements the log of the cross-nested logit model as a MEV model
    with the homogeneity parameters is explicitly involved.


    :param V: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.
    :type V: dict(int:biogeme.expressions.expr.Expression)

    :param availability: dict of objects representing the availability of each
               alternative, indexed
               by numerical ids. Must be consistent with V, or
               None. In this case, all alternatives are supposed to be
               always available.

    :type availability: dict(int:biogeme.expressions.expr.Expression)

    :param nests: a tuple containing as many items as nests. Each item is
        also a tuple containing two items

        - an object of type biogeme.expressions. expr.Expression representing
          the nest parameter,
        - a dictionary mapping the alternative ids with the cross-nested
          parameters for the corresponding nest. If an alternative is
          missing in the dictionary, the corresponding alpha is set to zero.

        Example::

            alphaA = {1: alpha1a,
                      2: alpha2a,
                      3: alpha3a,
                      4: alpha4a,
                      5: alpha5a,
                      6: alpha6a}
            alphaB = {1: alpha1b,
                      2: alpha2b,
                      3: alpha3b,
                      4: alpha4b,
                      5: alpha5b,
                      6: alpha6b}
            nesta = MUA , alphaA
            nestb = MUB , alphaB
            nests = nesta, nestb

    :type nests: tuple

    :param choice: id of the alternative for which the probability must be
              calculated.
    :type choice: biogeme.expressions.expr.Expression

    :param mu: Homogeneity parameter :math:`\\mu`.
    :type mu: biogeme.expressions.expr.Expression

    :param alone: list of alternatives that are alone in a nest
    :type alone: list(int)

    :return: log of the choice probability for the cross-nested logit model.
    :rtype: biogeme.expressions.expr.Expression

    :raise BiogemeError: if the definition of the nests is invalid.

    """
    ok, message = checkValidityCNL(V, nests, alone)
    if not ok:
        raise excep.BiogemeError(message)
    logGi = getMevForCrossNestedMu(V, availability, nests, mu, alone)
    logP = logmev(V, logGi, availability, choice)
    return logP


def checkValidityNestedLogit(V, nests, alone):
    """Verifies if the nested logit model is indeed based on a partition
    of the choice set.

    :param V: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.
    :type V: dict(int:biogeme.expressions.expr.Expression)
    :param nests: A tuple containing as many items as nests. Each item is also
        a tuple containing two items:

        - an object of type biogeme.expressions. expr.Expression
          representing the nest parameter,
        - a list containing the list of identifiers of the alternatives
          belonging to the nest.

        Example::

          nesta = MUA, [1, 2, 3]
          nestb = MUB, [4, 5, 6]
          nests = nesta, nestb

    :type nests: tuple

    :param alone: list of alternatives that are alone in a nest
    :type alone: list(int)

    :return: a tuple ok, message, where the message explains the
             problem is the nested structure is not OK.
    :rtype: tuple(bool, str)

    """

    ok = True
    message = ''

    fullChoiceSet = {i for i, v in V.items()}
    unionOfNests = set.union(*[set(n[1]) for n in nests])
    if alone is not None:
        unionOfNests = unionOfNests.union(alone)
    if fullChoiceSet != unionOfNests:
        ok = False
        d1 = fullChoiceSet.difference(unionOfNests)
        d2 = unionOfNests.difference(fullChoiceSet)
        if d1:
            message += (
                f'Alternative(s) in the choice set, but not in any nest:' f' {d1}\n'
            )
        if d2:
            message += (
                f'Alternative(s) in a nest, but not in the choice set: ' f'{d2}\n'
            )

    # Check that the set of alone alternatives is well defined
    if alone is not None:
        for n in nests:
            inter = set(alone).intersection(n[1])
            if inter:
                ok = False
                message += (
                    f'A nest contains the following alternative(s): '
                    f'{inter}, that is also declared alone in a nest.\n'
                )

    # Consider all pairs of nests and verify that the intersection is empty
    allPairs = [(n1, n2) for n1 in nests for n2 in nests if n1 != n2]
    for n1, n2 in allPairs:
        inter = set(n1[1]).intersection(n2[1])
        if inter:
            ok = False
            message += f'Two nests contain the following alternative(s): ' f'{inter}\n'

    if ok:
        message = 'The nested logit model is based on a partition. '

    return ok, message


def checkValidityCNL(V, nests, alone):
    """Verifies if the cross-nested logit specifciation is valid

    :param V: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.
    :type V: dict(int:biogeme.expressions.expr.Expression)

    :param nests: a tuple containing as many items as nests.
        Each item is also a tuple containing two items

        - an object of type biogeme.expressions.expr.Expression  representing
          the nest parameter,
        - a dictionary mapping the alternative ids with the cross-nested
          parameters for the corresponding nest. If an alternative is
          missing in the dictionary, the corresponding alpha is set to zero.

        Example::

            alphaA = {1: alpha1a,
                      2: alpha2a,
                      3: alpha3a,
                      4: alpha4a,
                      5: alpha5a,
                      6: alpha6a}
            alphaB = {1: alpha1b,
                      2: alpha2b,
                      3: alpha3b,
                      4: alpha4b,
                      5: alpha5b,
                      6: alpha6b}
            nesta = MUA, alphaA
            nestb = MUB, alphaB
            nests = nesta, nestb

    :type nests: tuple

    :param alone: list of alternatives that are alone in a nest
    :type alone: list(int)

    :return: a tuple ok, message, where the message explains the
             problem is the nested structure is not OK.
    :rtype: tuple(bool, str)
    """

    ok = True
    message = ''

    alt = {i: [] for i in V}
    number = 0
    for _, alpha in nests:
        for i, a in alpha.items():
            if a != 0.0:
                alt[i].append(a)
        number += 1

    problems_zero = []
    problems_one = []
    for i, ell in alt.items():
        if not ell:
            if alone is None or i not in alone:
                problems_zero.append(i)
                ok = False
        if len(ell) == 1 and isinstance(ell[0], expr.Expression):
            problems_one.append(i)

    if problems_zero:
        message += f'Alternative(s) not in any nest: {problems_zero}'

    if problems_one:
        message += (
            f' Alternative in exactly one nest, '
            f'and parameter alpha is defined by an '
            f'expression, and may not be constant: {problems_one}'
        )

    return ok, message


def ordered_likelihood(continuous_value, list_of_discrete_values, tau_parameter, cdf):
    """Ordered model that maps a continuous quantity with a list of
        discrete intervals (often logit or probit)

    Example: discrete values = [1, 2, 3, 4]

    We define thresholds tau_1_2, tau_2_3 and tau_3_4.
    In order to impose that the threshold are sorted, we actually define
        tau_1_2 = tau_parameter
        tau_2_3 = tau_1_2 + diff2
        tau_3_4 = tau_2_3 + diff3

    The probability that the discrete value is 2, say, is the
    probability that the continuous value lies between tau_1_2 and
    tau_2_3, where the probability distribution is logistic.

    :param continuous_value: continuous quantity to mapping
    :type continuous_value: biogeme.expressions.Expression

    :param list_of_discrete_values: discrete values
    :type list_of_discrete_values: list(int)

    :param tau_parameter: parameter for the first threshold
    :type tau_parameter: biogeme.expressions.Beta

    :param cdf: function calculating the CDF of the random variable
    :type cdf: fct(float) -> float

    :return: dict where the keys are the discrete values and the
        values are the corresponding probability.
    :rtype: dict(int: biogeme.expressions.Expression)
    """
    if not isinstance(tau_parameter, expr.Beta):
        error_msg = (
            f'tau_parameter must be a Beta expression, and not a {type(tau_parameter)}.'
        )
        raise excep.BiogemeError(error_msg)

    if len(list_of_discrete_values) == 2:
        the_proba = {
            list_of_discrete_values[0]: 1 - cdf(continuous_value - tau_parameter),
            list_of_discrete_values[1]: cdf(continuous_value - tau_parameter),
        }

        return the_proba

    diffs = {
        current_item: expr.Beta(
            f'{tau_parameter.name}_diff_{current_item}',
            1,
            0,
            None,
            0,
        )
        for current_item in list_of_discrete_values[1:-1]
    }

    # First term
    the_proba = {list_of_discrete_values[0]: 1 - cdf(continuous_value - tau_parameter)}

    # Intermediate terms
    tau = tau_parameter
    for item in list_of_discrete_values[1:-1]:
        next_tau = tau + diffs[item]
        the_proba[item] = cdf(continuous_value - tau) - cdf(continuous_value - next_tau)
        tau = next_tau

    # Last term
    the_proba[list_of_discrete_values[-1]] = cdf(continuous_value - tau)

    return the_proba


def ordered_logit(continuous_value, list_of_discrete_values, tau_parameter):
    """Ordered logit model that maps a continuous quantity with a
        list of discrete intervals

    Example: discrete values = [1, 2, 3, 4]

    We define thresholds tau_1_2, tau_2_3 and tau_3_4.
    In order to impose that the threshold are sorted, we actually define
        tau_1_2 = tau_parameter
        tau_2_3 = tau_1_2 + diff2
        tau_3_4 = tau_2_3 + diff3

    The probability that the discrete value is 2, say, is the
    probability that the continuous value lies between tau_1_2 and
    tau_2_3, where the probability distribution is logistic.

    :param continuous_value: continuous quantity to mapping
    :type continuous_value: biogeme.expressions.Expression

    :param list_of_discrete_values: discrete values
    :type list_of_discrete_values: list(int)

    :param tau_parameter: parameter for the first threshold
    :type tau_parameter: biogeme.expressions.Beta

    """
    return ordered_likelihood(
        continuous_value=continuous_value,
        list_of_discrete_values=list_of_discrete_values,
        tau_parameter=tau_parameter,
        cdf=dist.logisticcdf,
    )


def ordered_probit(continuous_value, list_of_discrete_values, tau_parameter):
    """Ordered probit model that maps a continuous quantity with a
        list of discrete intervals

    Example: discrete values = [1, 2, 3, 4]

    We define thresholds tau_1_2, tau_2_3 and tau_3_4.
    In order to impose that the threshold are sorted, we actually define
        tau_1_2 = tau_parameter
        tau_2_3 = tau_1_2 + diff2
        tau_3_4 = tau_2_3 + diff3

    The probability that the discrete value is 2, say, is the
    probability that the continuous value lies between tau_1_2 and
    tau_2_3, where the probability distribution is normal.

    :param continuous_value: continuous quantity to mapping
    :type continuous_value: biogeme.expressions.Expression

    :param list_of_discrete_values: discrete values
    :type list_of_discrete_values: list(int)

    :param tau_parameter: parameter for the first threshold
    :type tau_parameter: biogeme.expressions.Beta

    """
    return ordered_likelihood(
        continuous_value=continuous_value,
        list_of_discrete_values=list_of_discrete_values,
        tau_parameter=tau_parameter,
        cdf=expr.bioNormalCdf,
    )



"""File b01logit_wtp.py

:author: Michel Bierlaire, EPFL
:date: Fri Apr 14 15:23:14 2023

 Example of a logit model, where the utilities are expressed in the WTP space
"""

import biogeme.biogeme as bio
from biogeme import models
from biogeme.expressions import Beta

from swissmetro_data import (
    database,
    CHOICE,
    SM_AV,
    CAR_AV_SP,
    TRAIN_AV_SP,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    SM_TT_SCALED,
    SM_COST_SCALED,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
)

# Parameters to be estimated.
# In WTP space, the cost coefficient is normalized to -1, and the
# scale parameter MU is estimated
value_of_car = Beta('value_of_car', 0, None, None, 0)
value_of_train = Beta('value_of_train', 0, None, None, 0)
value_of_time = Beta('value_of_time', 0, None, None, 0)
mu = Beta('mu', 1, None, None, 0)


# Definition of the cost functions
cost_1 = TRAIN_COST_SCALED + value_of_train + value_of_time * TRAIN_TT_SCALED
cost_2 = SM_COST_SCALED + value_of_time * SM_TT_SCALED
cost_3 = CAR_CO_SCALED + value_of_car + value_of_time * CAR_TT_SCALED

# To obtain the utilities, we change the sign of the cost
# functions. We also scale them with the parameter mu

scaled_V1 = -mu * cost_1
scaled_V2 = -mu * cost_2
scaled_V3 = -mu * cost_3

# Associate utility functions with the numbering of alternatives
V = {1: scaled_V1, 2: scaled_V2, 3: scaled_V3}

# Associate the availability conditions with the alternatives
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# Definition of the model. This is the contribution of each
# observation to the log likelihood function.
logprob = models.loglogit(V, av, CHOICE)

# Create the Biogeme object
the_biogeme = bio.BIOGEME(database, logprob)
the_biogeme.modelName = 'b01logit_wtp'

# Calculate the null log likelihood for reporting.
the_biogeme.calculateNullLoglikelihood(av)

# Estimate the parameters
results = the_biogeme.estimate()
print(results.shortSummary())
# Get the results in a pandas table
pandas_results = results.getEstimatedParameters()
print(pandas_results)



"""Defines controllers for multiple expressions

:author: Michel Bierlaire
:date: Sun Jul 16 15:23:46 2023

"""
from itertools import product
from functools import partial, reduce
import random
import biogeme.exceptions as excep
from biogeme.parameters import biogeme_parameters
from biogeme.configuration import (
    SelectionTuple,
    Configuration,
    SELECTION_SEPARATOR,
    SEPARATOR,
)


class Controller:
    """Class controlling the specification of the Catalogs"""

    def __init__(self, controller_name, specification_names):
        """Constructor

        :param controller_name: name of the controller
        :type controller_name: str

        :param specification_names: list or tuple of the names of the
            specification controlled by the controller
        :type specification_names: list(str) or tuple(str)
        """
        self.controller_name = controller_name
        self.specification_names = specification_names
        self.current_index = 0
        self.dict_of_index = {
            name: index for index, name in enumerate(self.specification_names)
        }
        self.controlled_catalogs = []

    def all_configurations(self):
        """Return the code of all configurations

        :return: list of codes
        :rtype: list(str)
        """
        return [
            f'{self.controller_name}{SELECTION_SEPARATOR}{specification}'
            for specification in self.specification_names
        ]

    def __eq__(self, other):
        return self.controller_name == other.controller_name

    def __lt__(self, other):
        return self.controller_name < other.controller_name

    def __hash__(self):
        return hash(self.controller_name)

    def __repr__(self):
        return repr(self.controller_name)

    def __str__(self):
        return f'{self.controller_name}: {self.specification_names}'

    def controller_size(self):
        """Number of specifications managed by this controller"""
        return len(self.specification_names)

    def current_name(self):
        """Name of the currently selected expression"""
        return self.specification_names[self.current_index]

    def set_name(self, name):
        """Set the index of the controller based on the name of the specification

        :param name: name of the specification
        :type name: str
        """
        the_index = self.dict_of_index.get(name)
        if the_index is None:
            error_msg = (
                f'{name}: unknown specification for controller {self.controller_name}'
            )
            raise excep.BiogemeError(error_msg)
        self.set_index(the_index)

    def set_index(self, index):
        """Set the index of the controller, and update the controlled catalogs

        :param index: value of the index
        :type index: int

        :raises BiogemeError: if index is out of range
        """
        if index < 0 or index >= self.controller_size():
            error_msg = (
                f'Wrong index {index} for controller {self.controller_name}. '
                f'Must be in [0, {self.controller_size()}]'
            )
            raise excep.BiogemeError(error_msg)

        self.current_index = index
        for catalog in self.controlled_catalogs:
            catalog.current_index = index

    def reset_selection(self):
        """Select the first specification"""
        self.set_index(0)

    def modify_controller(self, step, circular):
        """Modify the specification of the controller

        :param step: increment of the modifications. Can be negative.
        :type step: int

        :param circular: If True, the modificiation is always made. If
            the selection needs to move past the last one, it comes
            back to the first one. For instance, if the catalog is
            currently at its last value, and the step is 1, it is set
            to its first value. If circular is False, and the
            selection needs to move past the last one, the selection
            is set to the last one. It works symmetrically if the step
            is negative
        :type circular: bool

        :return: number of actual modifications
        :rtype: int

        """
        the_size = self.controller_size()
        new_index = self.current_index + step
        if circular:
            self.set_index(new_index % the_size)
            return step

        if new_index < 0:
            total_modif = self.current_index
            self.set_index(0)
            return total_modif

        if new_index >= the_size:
            total_modif = the_size - 1 - self.current_index
            self.set_index(the_size - 1)
            return total_modif

        self.set_index(new_index)
        return step


class CentralController:
    """Class controlling the complete multiple expression"""

    def __init__(self, expression):
        """Constructor

        :param controllers: all controllers of the expression
        :type controllers: tuple(controllers)

        :param expression: controllers expression
        :type expression: biogeme.expressions.Expression
        """

        set_of_controllers = expression.get_all_controllers()

        self.controllers = tuple(sorted(set_of_controllers))
        self.dict_of_controllers = {
            controller.controller_name: controller for controller in self.controllers
        }
        self.expression = expression

        all_controllers_states = [
            the_controller.all_configurations() for the_controller in self.controllers
        ]

        # The total number of configurations is the product of the
        # length of each controller.

        if all_controllers_states:
            self._number_of_configurations = reduce(
                lambda x, y: x * y, map(len, all_controllers_states)
            )
        else:
            self._number_of_configurations = 0

        maximum_number_of_configurations = biogeme_parameters.get_value(
            name='maximum_number_catalog_expressions', section='Estimation'
        )
        if self.number_of_configurations() > maximum_number_of_configurations:
            self.all_configurations_ids = None
            self.all_configurations = None
            return

        self.all_configurations_ids = {
            SEPARATOR.join(combination)
            for combination in product(*all_controllers_states)
        }
        self.all_configurations = {
            Configuration.from_string(id) for id in self.all_configurations_ids
        }

    def number_of_configurations(self):
        """Total number of configurations"""
        return self._number_of_configurations

    def get_configuration(self):
        """Obtain the current configuration of the controllers"""
        selections = (
            SelectionTuple(
                controller=controller.controller_name,
                selection=controller.current_name(),
            )
            for controller in self.controllers
        )
        return Configuration(selections)

    def set_configuration(self, configuration):
        """Apply a configuration to the controllers

        :param configuration: the configuration to be applied
        :type configuration: Configuration
        """
        properly_set = {
            controller.controller_name: False for controller in self.controllers
        }
        for selection in configuration.selections:
            controller = self.dict_of_controllers.get(selection.controller)
            if controller is None:
                error_msg = (
                    f'Wrong configuration: {configuration}. Controller '
                    f'{selection.controller} is unknown'
                )
                raise excep.BiogemeError(error_msg)
            controller.set_name(selection.selection)
            properly_set[selection.controller] = True

        missing_controllers = [name for name, done in properly_set.items() if not done]
        if missing_controllers:
            error_msg = (
                f'Incomplete configuration {configuration}. The following controllers '
                f'are not defined: {missing_controllers}'
            )
            raise excep.BiogemeError(error_msg)

    def set_configuration_from_id(self, configuration_id):
        """Apply a configuration to the controllers

        :param configuration_id: the ID of the configuration to be applied
        :type configuration_id: str
        """
        the_configuration = Configuration.from_string(configuration_id)
        self.set_configuration(the_configuration)

    def set_controller(self, controller_name, index):
        """Set the given controller to the specified index

        :param controller_name: name of the controller
        :type controller_name: string

        :param index: index of the selection
        :type index: int
        """
        the_controller = self.dict_of_controllers.get(controller_name)

        if the_controller is None:
            error_msg = f'Unknown controller: {controller_name}'
            raise excep.BiogemeError(error_msg)

        the_controller.set_index(index)

    def increased_controller(self, controller_name, current_config, step):
        """Increase the selection of one controller by "step"

        :param controller_name: name of the controller to modify
        :type controller_name: str

        :param current_config: current configuration
        :type current_config: str

        :param step: number of steps to perform
        :type step: int
        """
        self.set_configuration_from_id(current_config)
        the_controller = self.dict_of_controllers.get(controller_name)
        if the_controller is None:
            error_msg = f'Unknown controller {the_controller}'
            raise excep.BiogemeError(error_msg)
        the_controller.modify_controller(step=step, circular=True)
        new_config = self.get_configuration()
        return new_config, step

    def decreased_controller(self, controller_name, current_config, step):
        """Decrease the selection of one controller by "step"

        :param controller_name: name of the controller to modify
        :type controller_name: str

        :param current_config: current configuration
        :type current_config: str

        :param step: number of steps to perform
        :type step: int

        :return: ID of the new configuration and number of steps performed.
        :rtype: tuple(str, int)
        """
        self.set_configuration_from_id(current_config)
        the_controller = self.dict_of_controllers.get(controller_name)
        if the_controller is None:
            error_msg = f'Unknown controller {the_controller}'
            raise excep.BiogemeError(error_msg)
        the_controller.modify_controller(step=-step, circular=True)
        new_config = self.get_configuration()
        return new_config, step

    def two_controllers(
        self,
        first_controller_name,
        second_controller_name,
        direction,
        current_config,
        step,
    ):
        """Modification of two controllers. Meaning of direction:

        - NE (North-East): first controller increased by "step", second
              controller increased by "step"
        - NW (North-West): first controller decreased by "step", second
              controller increased by "step"
        - SE (South-East): first controller increased by "step", second
              controller decreased by "step"
        - SW (South-West): first controller decreased by "step", second
              controller decreased by "step"

        :param first_controller_name: name of the first_controller to modify
        :type first_controller_name: str

        :param second_controller_name: name of the second_controller to modify
        :type second_controller_name: str

        :param direction: direction based on the compass rose. Must be NE, NW, SE or SW
        :type direction: str

        :param current_config: current configuration
        :type current_config: str

        :param step: number of steps to perform
        :type step: int

        :return: ID of the new configuration and number of steps performed.
        :rtype: tuple(str, int)

        """
        if direction not in ['NE', 'NW', 'SE', 'SW']:
            error_msg = f'Incorrect direction {direction}. Must be NE, NW, SE or SW'
            raise excep.BiogemeError(error_msg)
        self.set_configuration_from_id(current_config)
        the_first_controller = self.dict_of_controllers.get(first_controller_name)
        if the_first_controller is None:
            error_msg = f'Unknown controller {the_first_controller}'
            raise excep.BiogemeError(error_msg)
        the_second_controller = self.dict_of_controllers.get(second_controller_name)
        if the_second_controller is None:
            error_msg = f'Unknown controller {the_second_controller}'
            raise excep.BiogemeError(error_msg)
        # The direction for the first controller is W (decrease) or E (increase)
        # The direction for the second controller is N (increase) or S (decrease)
        actual_step = step if direction[1] == 'E' else -step
        the_first_controller.modify_controller(step=actual_step, circular=True)
        actual_step = step if direction[0] == 'N' else -step
        the_second_controller.modify_controller(step=actual_step, circular=True)
        new_config = self.get_configuration()
        return new_config, step

    def modify_random_controllers(
        self,
        increase,
        current_config,
        step,
    ):
        """Increase the selection of "step" controllers by 1

        :param increase: If True, the indices are increased . If
        False, they are decreased.
        :type increase: bool

        :param current_config: current configuration
        :type current_config: str

        :param step: number of steps to perform
        :type step: int

        """
        self.set_configuration_from_id(current_config)
        number_of_controllers = len(self.controllers)
        actual_size = min(step, number_of_controllers)
        selected_controllers = random.choices(
            list(self.dict_of_controllers.keys()), k=actual_size
        )
        the_modification = 1 if increase else 1
        for the_controller in selected_controllers:
            self.dict_of_controllers[the_controller].modify_controller(
                step=the_modification, circular=True
            )
        new_config = self.get_configuration()
        return new_config, actual_size

    def prepare_operators(self):
        """Operators are functions that take a configuration and a
        size as arguments, and return a new configuration, and the
        actual number of modifications that have been
        implemented.

        """

        dict_of_operators = {}
        # Increase and decrease controllers
        for name in self.dict_of_controllers.keys():
            dict_of_operators[f'Increase {name}'] = partial(
                self.increased_controller, controller_name=name
            )
            dict_of_operators[f'Decrease {name}'] = partial(
                self.decreased_controller, controller_name=name
            )
        # Pair of controllers
        directions = ['NE', 'NW', 'SE', 'SW']
        # Using nested list comprehensions to generate all possible pairs
        all_pairs = [
            (name1, name2, direction)
            for name1 in self.dict_of_controllers.keys()
            for name2 in self.dict_of_controllers.keys()
            for direction in directions
        ]

        # Filter out pairs with the same name
        filtered_pairs = [
            (name1, name2, direction)
            for name1, name2, direction in all_pairs
            if name1 != name2
        ]

        for name1, name2, direction in filtered_pairs:
            dict_of_operators[f'Pair_{name1}_{name2}_{direction}'] = partial(
                self.two_controllers,
                first_controller_name=name1,
                second_controller_name=name2,
                direction=direction,
            )

        # Several controllers
        dict_of_operators['Increase_several'] = partial(
            self.modify_random_controllers, increase=True
        )
        dict_of_operators['Decrease_several'] = partial(
            self.modify_random_controllers, increase=False
        )
        return dict_of_operators



"""File b01model.py

:author: Michel Bierlaire, EPFL
:date: Fri Jul 14 09:47:21 2023

Investigate several choice models:
- logit
- nested logit with two nests: public and private transportation
- nested logit with two nests existing and future modes
for a total of 3 specifications.
"""
import biogeme.biogeme as bio
from biogeme import models
from biogeme.expressions import Beta
from biogeme.catalog import Catalog
from results_analysis import report
from swissmetro_data import (
    database,
    CHOICE,
    SM_AV,
    CAR_AV_SP,
    TRAIN_AV_SP,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    SM_TT_SCALED,
    SM_COST_SCALED,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
)

# Parameters to be estimated
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
B_TIME = Beta('B_TIME', 0, None, None, 0)
B_COST = Beta('B_COST', 0, None, None, 0)


# Definition of the utility functions
V1 = ASC_TRAIN + B_TIME * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
V2 = B_TIME * SM_TT_SCALED + B_COST * SM_COST_SCALED
V3 = ASC_CAR + B_TIME * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# Associate utility functions with the numbering of alternatives
V = {1: V1, 2: V2, 3: V3}

# Associate the availability conditions with the alternatives
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# Definition of the model. This is the contribution of each
# observation to the log likelihood function.
logprob_logit = models.loglogit(V, av, CHOICE)

MU_existing = Beta('MU_existing', 1, 1, 10, 0)
existing = MU_existing, [1, 3]
future = 1.0, [2]
nests_existing = existing, future
logprob_nested_existing = models.lognested(V, av, nests_existing, CHOICE)

MU_public = Beta('MU_public', 1, 1, 10, 0)
public = MU_public, [1, 2]
private = 1.0, [3]
nests_public = public, private
logprob_nested_public = models.lognested(V, av, nests_public, CHOICE)

model_catalog = Catalog.from_dict(
    catalog_name='model_catalog',
    dict_of_expressions={
        'logit': logprob_logit,
        'nested existing': logprob_nested_existing,
        'nested public': logprob_nested_public,
    },
)
# Create the Biogeme object
the_biogeme = bio.BIOGEME(database, model_catalog)
the_biogeme.modelName = 'b05model'
the_biogeme.generate_html = False
the_biogeme.generate_pickle = False

# Estimate the parameters
dict_of_results = the_biogeme.estimate_catalog()

report(dict_of_results)



"""File b14nested_endogenous_sampling.py

:author: Michel Bierlaire, EPFL
:date: Sun Apr  9 18:25:03 2023

 Example of a nested logit model, with the corrections for endogenous sampling.
 Three alternatives: Train, Car and Swissmetro
 Train and car are in the same nest.

"""

import numpy as np
import biogeme.logging as blog
import biogeme.biogeme as bio
from biogeme import models
from biogeme.expressions import Beta

from swissmetro_data import (
    database,
    CHOICE,
    SM_AV,
    CAR_AV_SP,
    TRAIN_AV_SP,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    SM_TT_SCALED,
    SM_COST_SCALED,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
)

logger = blog.get_screen_logger(level=blog.INFO)
logger.info('Example b14nested_endogenous_sampling.py')

# Parameters to be estimated
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
ASC_SM = Beta('ASC_SM', 0, None, None, 1)
B_TIME = Beta('B_TIME', 0, None, None, 0)
B_COST = Beta('B_COST', 0, None, None, 0)
MU = Beta('MU', 1, 1, 10, 0)

# In this example, we assume that the three modes exist, and that the
# sampling protocal is choice-based. The probability that a respondent
# belongs to the sample is R_i.
R_TRAIN = 4.42e-2
R_SM = 3.36e-3
R_CAR = 7.5e-3

# The correction terms are the log of these quantities
correction = {1: np.log(R_TRAIN), 2: np.log(R_SM), 3: np.log(R_CAR)}

# Definition of the utility functions
V1 = ASC_TRAIN + B_TIME * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
V2 = ASC_SM + B_TIME * SM_TT_SCALED + B_COST * SM_COST_SCALED
V3 = ASC_CAR + B_TIME * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# Associate utility functions with the numbering of alternatives
V = {1: V1, 2: V2, 3: V3}

# Associate the availability conditions with the alternatives
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# Definition of nests:
# 1: nests parameter
# 2: list of alternatives
existing = MU, [1, 3]
future = 1.0, [2]
nests = existing, future

# The choice model is a nested logit, with corrections for endogenous sampling
# We first obtain the expression of the Gi function for nested logit.
Gi = models.getMevForNested(V, av, nests)

# Then we calculate the MEV log probability, accounting for the correction.
logprob = models.logmev_endogenousSampling(V, Gi, av, correction, CHOICE)

# Create the Biogeme object
the_biogeme = bio.BIOGEME(database, logprob)
the_biogeme.modelName = 'b14nested_endogenous_eampling'

# Estimate the parameters
results = the_biogeme.estimate()
print(results.short_summary())
pandas_results = results.getEstimatedParameters()
print(pandas_results)



"""File b04market_shares.py

:author: Michel Bierlaire, EPFL
:date: Wed Apr 12 21:03:52 2023

We use an estimated model to calcualte market shares.
"""
import sys
from biogeme import models
import biogeme.biogeme as bio
import biogeme.exceptions as excep
import biogeme.results as res
from optima_data import database, normalized_weight
from scenarios import scenario

# Obtain the specification for the default scenario
V, nests, _, _ = scenario()

# Obtain the expression for the choice probability of each alternative
prob_PT = models.nested(V, None, nests, 0)
prob_CAR = models.nested(V, None, nests, 1)
prob_SM = models.nested(V, None, nests, 2)

# Read the estimation results from the file
try:
    results = res.bioResults(pickleFile='b02estimation.pickle')
except excep.BiogemeError:
    sys.exit(
        'Run first the script b02simulation.py '
        'in order to generate the '
        'file b02estimation.pickle.'
    )

# We now simulate the choice probabilities and the weight

simulate = {
    'weight': normalized_weight,
    'Prob. PT': prob_PT,
    'Prob. car': prob_CAR,
    'Prob. SM': prob_SM,
}

the_biogeme = bio.BIOGEME(database, simulate)
simulated_values = the_biogeme.simulate(results.getBetaValues())

# We also calculate confidence intervals for the calculated quantities

betas = the_biogeme.freeBetaNames()
b = results.getBetasForSensitivityAnalysis(betas)
left, right = the_biogeme.confidenceIntervals(b, 0.9)


# Market shares are calculated using the weighted mean of the
# individual probabilities

# Alternative car
simulated_values['Weighted prob. car'] = (
    simulated_values['weight'] * simulated_values['Prob. car']
)
left['Weighted prob. car'] = left['weight'] * left['Prob. car']
right['Weighted prob. car'] = right['weight'] * right['Prob. car']

marketShare_car = simulated_values['Weighted prob. car'].mean()
marketShare_car_left = left['Weighted prob. car'].mean()
marketShare_car_right = right['Weighted prob. car'].mean()

# Alternative public transportation
simulated_values['Weighted prob. PT'] = (
    simulated_values['weight'] * simulated_values['Prob. PT']
)
left['Weighted prob. PT'] = left['weight'] * left['Prob. PT']
right['Weighted prob. PT'] = right['weight'] * right['Prob. PT']

marketShare_PT = simulated_values['Weighted prob. PT'].mean()
marketShare_PT_left = left['Weighted prob. PT'].mean()
marketShare_PT_right = right['Weighted prob. PT'].mean()

# Alternative slow modes
simulated_values['Weighted prob. SM'] = (
    simulated_values['weight'] * simulated_values['Prob. SM']
)
left['Weighted prob. SM'] = left['weight'] * left['Prob. SM']
right['Weighted prob. SM'] = right['weight'] * right['Prob. SM']

marketShare_SM = simulated_values['Weighted prob. SM'].mean()
marketShare_SM_left = left['Weighted prob. SM'].mean()
marketShare_SM_right = right['Weighted prob. SM'].mean()

# Reporting
print(
    f'Market share for car: {100*marketShare_car:.1f}% '
    f'[{100*marketShare_car_left:.1f}%, '
    f'{100*marketShare_car_right:.1f}%]'
)
print(
    f'Market share for PT:  {100*marketShare_PT:.1f}% '
    f'[{100*marketShare_PT_left:.1f}%, '
    f'{100*marketShare_PT_right:.1f}%]'
)
print(
    f'Market share for SM:   {100*marketShare_SM:.1f}% '
    f'[{100*marketShare_SM_left:.1f}%, '
    f'{100*marketShare_SM_right:.1f}%]'
)



"""File b23binary_probit.py

:author: Michel Bierlaire, EPFL
:date: Wed Apr 12 17:58:18 2023

 Example of a binary probit model.
 Two alternatives: Train and Car
"""

import biogeme.biogeme as bio
from biogeme.expressions import Beta, bioNormalCdf, Elem, log
from swissmetro_binary import (
    database,
    CHOICE,
    TRAIN_AV_SP,
    CAR_AV_SP,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
)

# Parameters to be estimated
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
B_TIME_CAR = Beta('B_TIME_CAR', 0, None, None, 0)
B_TIME_TRAIN = Beta('B_TIME_TRAIN', 0, None, None, 0)
B_COST_CAR = Beta('B_COST_CAR', 0, None, None, 0)
B_COST_TRAIN = Beta('B_COST_TRAIN', 0, None, None, 0)

# Definition of the utility functions
# We estimate a binary probit model. There are only two alternatives.
V1 = B_TIME_TRAIN * TRAIN_TT_SCALED + B_COST_TRAIN * TRAIN_COST_SCALED
V3 = ASC_CAR + B_TIME_CAR * CAR_TT_SCALED + B_COST_CAR * CAR_CO_SCALED

# Associate choice probability with the numbering of alternatives
# If one alternative is not available, the choice probability of the other one is 1.
logP = {
    1: TRAIN_AV_SP * (CAR_AV_SP * log(bioNormalCdf(V1 - V3) + 1 - CAR_AV_SP)),
    3: CAR_AV_SP * (TRAIN_AV_SP * log(bioNormalCdf(V3 - V1) + 1 - TRAIN_AV_SP)),
}

# Definition of the model. This is the contribution of each
# observation to the log likelihood function.
logprob = Elem(logP, CHOICE)

# Create the Biogeme object
the_biogeme = bio.BIOGEME(database, logprob)
the_biogeme.modelName = 'b23probit'

# Estimate the parameters
results = the_biogeme.estimate()
print(results.short_summary())
pandas_results = results.getEstimatedParameters()
print(pandas_results)



"""File b11cnl_simul.py

:author: Michel Bierlaire, EPFL
:date: Sun Apr  9 18:08:29 2023

 Example of simulation with a cross-nested logit model.
 Three alternatives: Train, Car and Swissmetro
 Train and car are in the same nest.
"""

import sys
import biogeme.biogeme as bio
from biogeme import models
from biogeme.results import calculate_correlation
import biogeme.results as res
from biogeme.expressions import Beta, Derive
import biogeme.exceptions as excep
from swissmetro_data import (
    database,
    SM_AV,
    CAR_AV_SP,
    TRAIN_AV_SP,
    TRAIN_TT_SCALED,
    TRAIN_TT,
    TRAIN_COST_SCALED,
    SM_TT_SCALED,
    SM_TT,
    SM_COST_SCALED,
    CAR_TT_SCALED,
    CAR_TT,
    CAR_CO_SCALED,
)

# Simulation should be done with estimated value of the
# parameters. You can include them manually. Here, we prefer to set
# them to zero, and read the values from the file created after
# estimation.
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
ASC_SM = Beta('ASC_SM', 0, None, None, 1)
B_TIME = Beta('B_TIME', 0, None, None, 0)
B_COST = Beta('B_COST', 0, None, None, 0)

MU_EXISTING = Beta('MU_EXISTING', 1, 1, None, 0)
MU_PUBLIC = Beta('MU_PUBLIC', 1, 1, None, 0)
ALPHA_EXISTING = Beta('ALPHA_EXISTING', 0.5, 0, 1, 0)
ALPHA_PUBLIC = 1 - ALPHA_EXISTING

# Definition of the utility functions
V1 = ASC_TRAIN + B_TIME * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
V2 = ASC_SM + B_TIME * SM_TT_SCALED + B_COST * SM_COST_SCALED
V3 = ASC_CAR + B_TIME * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# Associate utility functions with the numbering of alternatives
V = {1: V1, 2: V2, 3: V3}

# Associate the availability conditions with the alternatives
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}
# Definition of nests
# Nest membership parameters
alpha_existing = {1: ALPHA_EXISTING, 2: 0.0, 3: 1.0}

alpha_public = {1: ALPHA_PUBLIC, 2: 1.0, 3: 0.0}

nest_existing = MU_EXISTING, alpha_existing
nest_public = MU_PUBLIC, alpha_public
nests = nest_existing, nest_public

# Instead of estimating the parameters, read the estimation
# results from the pickle file.
try:
    results = res.bioResults(pickleFile='b11cnl.pickle')
except excep.BiogemeError:
    print('Run first the script b11cnl.py in order to generate the file 11cnl.pickle.')
    sys.exit()

print('Estimation results: ', results.getEstimatedParameters())

print('Calculating correlation matrix. It may generate numerical warnings from scipy.')
corr = calculate_correlation(
    nests, results, alternative_names={1: 'Train', 2: 'Swissmetro', 3: 'Car'}
)
print(corr)

# The choice model is a cross-nested logit, with availability conditions
prob1 = models.cnl_avail(V, av, nests, 1)
prob2 = models.cnl_avail(V, av, nests, 2)
prob3 = models.cnl_avail(V, av, nests, 3)

# We calculate elasticities
genelas1 = Derive(prob1, 'TRAIN_TT') * TRAIN_TT / prob1
genelas2 = Derive(prob2, 'SM_TT') * SM_TT / prob2
genelas3 = Derive(prob3, 'CAR_TT') * CAR_TT / prob3

# We report the probability of each alternative and the elasticities
simulate = {
    'Prob. train': prob1,
    'Prob. Swissmetro': prob2,
    'Prob. car': prob3,
    'Elas. 1': genelas1,
    'Elas. 2': genelas2,
    'Elas. 3': genelas3,
}

# Create the Biogeme object
biosim = bio.BIOGEME(database, simulate)
biosim.modelName = 'b11cnl_simul'

# Perform the simulation
simresults = biosim.simulate(results.getBetaValues())
print('Simulation results')
print(simresults.describe())

print(f'Aggregate share of train: {100*simresults["Prob. train"].mean():.1f}%')



import os
import unittest
import pandas as pd
import biogeme.database as db
import biogeme.biogeme as bio
from biogeme.expressions import Beta, bioNormalCdf, Elem, log


myPath = os.path.dirname(os.path.abspath(__file__))
df = pd.read_csv(f'{myPath}/swissmetro.dat', sep='\t')
database = db.Database('swissmetro', df)

# The Pandas data structure is available as database.data. Use all the
# Pandas functions to invesigate the database
# print(database.data.describe())

globals().update(database.variables)

# As we estimate a binary model, we remove observations where
# Swissmetro was chosen (CHOICE == 2). We also remove observations
# where one of the two alternatives is not available.

CAR_AV_SP = database.DefineVariable('CAR_AV_SP', CAR_AV * (SP != 0))
TRAIN_AV_SP = database.DefineVariable('TRAIN_AV_SP', TRAIN_AV * (SP != 0))
exclude = (TRAIN_AV_SP == 0) + (CAR_AV_SP == 0) + (CHOICE == 2) + (
    (PURPOSE != 1) * (PURPOSE != 3) + (CHOICE == 0)
) > 0


database.remove(exclude)


ASC_CAR = Beta('ASC_CAR', 1, None, None, 0)
B_TIME = Beta('B_TIME', 1, None, None, 0)
B_COST = Beta('B_COST', 1, None, None, 0)


TRAIN_COST = TRAIN_CO * (GA == 0)

TRAIN_TT_SCALED = database.DefineVariable('TRAIN_TT_SCALED', TRAIN_TT / 100.0)
TRAIN_COST_SCALED = database.DefineVariable(
    'TRAIN_COST_SCALED', TRAIN_COST / 100
)
CAR_TT_SCALED = database.DefineVariable('CAR_TT_SCALED', CAR_TT / 100)
CAR_CO_SCALED = database.DefineVariable('CAR_CO_SCALED', CAR_CO / 100)

# We estimate a binary probit model. There are only two alternatives.
V1 = B_TIME * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
V3 = ASC_CAR + B_TIME * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# Associate choice probability with the numbering of alternatives

P = {1: bioNormalCdf(V1 - V3), 3: bioNormalCdf(V3 - V1)}


prob = Elem(P, CHOICE)


class test_02(unittest.TestCase):
    def testEstimation(self):
        biogeme = bio.BIOGEME(database, log(prob))
        biogeme.saveIterations = False
        biogeme.generateHtml = False
        biogeme.generatePickle = False
        results = biogeme.estimate()
        self.assertAlmostEqual(results.data.logLike, -986.1888, 2)


if __name__ == '__main__':
    unittest.main()



"""File airline_spec.py

:author: Michel Bierlaire, EPFL
:date: Sun Apr  9 16:22:32 2023

Specification for the assisted specification algorithm. Airline case study
"""


from biogeme import models
import biogeme.biogeme as bio
from biogeme.catalog import (
    Catalog,
    segmentation_catalogs,
    generic_alt_specific_catalogs,
)
from biogeme.expressions import (
    Beta,
    log,
    logzero,
    Variable,
)

from airline_data import (
    database,
    chosenAlternative,
    Fare_1,
    Fare_2,
    Fare_3,
    Legroom_1,
    Legroom_2,
    Legroom_3,
    TripTimeHours_1,
    TripTimeHours_2,
    TripTimeHours_3,
    Opt1_SchedDelayEarly,
    Opt2_SchedDelayEarly,
    Opt3_SchedDelayEarly,
    Opt1_SchedDelayLate,
    Opt2_SchedDelayLate,
    Opt3_SchedDelayLate,
    q02_TripPurpose,
    q03_WhoPays,
    q17_Gender,
    q11_DepartureOrArrivalIsImportant,
    education,
)

# Define the catalog for the time variable

ell_fare = Beta('lambda_fare', 1, None, 10, 0)

fare_direct_catalog = Catalog.from_dict(
    catalog_name='fare_direct_catalog',
    dict_of_expressions={
        'linear': Fare_1,
        'income_interaction_1': Fare_1 / Variable('Cont_Income'),
        'income_interaction_2': Fare_1 + Fare_1 / Variable('Cont_Income'),
        'log_income_interaction': log(Fare_1) / Variable('Cont_Income'),
        'sqrt_income_interaction': Fare_1**0.5 / Variable('Cont_Income'),
        'log': logzero(Fare_1),
        'sqrt': Fare_1**0.5,
        'square': Fare_1 * Fare_1,
        'boxcox': models.boxcox(Fare_1, ell_fare),
    },
)

fare_same_catalog = Catalog.from_dict(
    catalog_name='fare_same_catalog',
    dict_of_expressions={
        'linear': Fare_2,
        'income_interaction_1': Fare_2 / Variable('Cont_Income'),
        'income_interaction_2': Fare_2 + Fare_2 / Variable('Cont_Income'),
        'log_income_interaction': log(Fare_2) / Variable('Cont_Income'),
        'sqrt_income_interaction': Fare_2**0.5 / Variable('Cont_Income'),
        'log': logzero(Fare_2),
        'sqrt': Fare_2**0.5,
        'square': Fare_2 * Fare_2,
        'boxcox': models.boxcox(Fare_2, ell_fare),
    },
    controlled_by=fare_direct_catalog.controlled_by,
)

fare_multiple_catalog = Catalog.from_dict(
    catalog_name='fare_multiple_catalog',
    dict_of_expressions={
        'linear': Fare_3,
        'income_interaction_1': Fare_3 / Variable('Cont_Income'),
        'income_interaction_2': Fare_3 + Fare_3 / Variable('Cont_Income'),
        'log_income_interaction': log(Fare_3) / Variable('Cont_Income'),
        'sqrt_income_interaction': Fare_3**0.5 / Variable('Cont_Income'),
        'log': logzero(Fare_3),
        'sqrt': Fare_3**0.5,
        'square': Fare_3 * Fare_3,
        'boxcox': models.boxcox(Fare_3, ell_fare),
    },
    controlled_by=fare_direct_catalog.controlled_by,
)

legroom_direct_catalog = Catalog.from_dict(
    catalog_name='legroom_direct_catalog',
    dict_of_expressions={
        'linear': Legroom_1,
        'log': logzero(Legroom_1),
        'sqrt': Legroom_1**0.5,
        'square': Legroom_1 * Legroom_1,
    },
)

legroom_same_catalog = Catalog.from_dict(
    catalog_name='legroom_same_catalog',
    dict_of_expressions={
        'linear': Legroom_2,
        'log': logzero(Legroom_2),
        'sqrt': Legroom_2**0.5,
        'square': Legroom_2 * Legroom_2,
    },
    controlled_by=legroom_direct_catalog.controlled_by,
)

legroom_multiple_catalog = Catalog.from_dict(
    catalog_name='legroom_multiple_catalog',
    dict_of_expressions={
        'linear': Legroom_3,
        'log': logzero(Legroom_3),
        'sqrt': Legroom_3**0.5,
        'square': Legroom_3 * Legroom_3,
    },
    controlled_by=legroom_direct_catalog.controlled_by,
)

ell_time = Beta('lambda_time', 1, None, 10, 0)

time_direct_catalog = Catalog.from_dict(
    catalog_name='time_direct_catalog',
    dict_of_expressions={
        'linear': TripTimeHours_1,
        'log': logzero(TripTimeHours_1),
        'sqrt': TripTimeHours_1**0.5,
        'square': TripTimeHours_1 * TripTimeHours_1,
        'boxcox': models.boxcox(TripTimeHours_1, ell_time),
        'piecewise_1': models.piecewise_as_variable(
            TripTimeHours_1, [0, 2, 4, 6, 8, None]
        ),
        'piecewise_2': models.piecewise_as_variable(TripTimeHours_1, [0, 2, 8, None]),
    },
)

time_same_catalog = Catalog.from_dict(
    catalog_name='time_same_catalog',
    dict_of_expressions={
        'linear': TripTimeHours_2,
        'log': logzero(TripTimeHours_2),
        'sqrt': TripTimeHours_2**0.5,
        'square': TripTimeHours_2 * TripTimeHours_2,
        'boxcox': models.boxcox(TripTimeHours_2, ell_time),
        'piecewise_1': models.piecewise_as_variable(
            TripTimeHours_2, [0, 2, 4, 6, 8, None]
        ),
        'piecewise_2': models.piecewise_as_variable(TripTimeHours_2, [0, 2, 8, None]),
    },
    controlled_by=time_direct_catalog.controlled_by,
)

time_multiple_catalog = Catalog.from_dict(
    catalog_name='time_multiple_catalog',
    dict_of_expressions={
        'linear': TripTimeHours_3,
        'log': logzero(TripTimeHours_3),
        'sqrt': TripTimeHours_3**0.5,
        'square': TripTimeHours_3 * TripTimeHours_3,
        'boxcox': models.boxcox(TripTimeHours_3, ell_time),
        'piecewise_1': models.piecewise_as_variable(
            TripTimeHours_3, [0, 2, 4, 6, 8, None]
        ),
        'piecewise_2': models.piecewise_as_variable(TripTimeHours_3, [0, 2, 8, None]),
    },
    controlled_by=time_direct_catalog.controlled_by,
)

early_direct_catalog = Catalog.from_dict(
    catalog_name='early_direct_catalog',
    dict_of_expressions={
        'linear': Opt1_SchedDelayEarly,
        'log': logzero(Opt1_SchedDelayEarly),
        'sqrt': Opt1_SchedDelayEarly**0.5,
        'square': Opt1_SchedDelayEarly * Opt1_SchedDelayEarly,
    },
)

early_same_catalog = Catalog.from_dict(
    catalog_name='early_same_catalog',
    dict_of_expressions={
        'linear': Opt2_SchedDelayEarly,
        'log': logzero(Opt2_SchedDelayEarly),
        'sqrt': Opt2_SchedDelayEarly**0.5,
        'square': Opt2_SchedDelayEarly * Opt2_SchedDelayEarly,
    },
    controlled_by=early_direct_catalog.controlled_by,
)

early_multiple_catalog = Catalog.from_dict(
    catalog_name='early_multiple_catalog',
    dict_of_expressions={
        'linear': Opt3_SchedDelayEarly,
        'log': logzero(Opt3_SchedDelayEarly),
        'sqrt': Opt3_SchedDelayEarly**0.5,
        'square': Opt3_SchedDelayEarly * Opt3_SchedDelayEarly,
    },
    controlled_by=early_direct_catalog.controlled_by,
)

late_direct_catalog = Catalog.from_dict(
    catalog_name='late_direct_catalog',
    dict_of_expressions={
        'linear': Opt1_SchedDelayLate,
        'log': logzero(Opt1_SchedDelayLate),
        'sqrt': Opt1_SchedDelayLate**0.5,
        'square': Opt1_SchedDelayLate * Opt1_SchedDelayLate,
    },
    controlled_by=early_direct_catalog.controlled_by,
)

late_same_catalog = Catalog.from_dict(
    catalog_name='late_same_catalog',
    dict_of_expressions={
        'linear': Opt2_SchedDelayLate,
        'log': logzero(Opt2_SchedDelayLate),
        'sqrt': Opt2_SchedDelayLate**0.5,
        'square': Opt2_SchedDelayLate * Opt2_SchedDelayLate,
    },
    controlled_by=early_direct_catalog.controlled_by,
)

late_multiple_catalog = Catalog.from_dict(
    catalog_name='late_multiple_catalog',
    dict_of_expressions={
        'linear': Opt3_SchedDelayLate,
        'log': logzero(Opt3_SchedDelayLate),
        'sqrt': Opt3_SchedDelayLate**0.5,
        'square': Opt3_SchedDelayLate * Opt3_SchedDelayLate,
    },
    controlled_by=early_direct_catalog.controlled_by,
)

# Define the potential segmentations
all_segmentations = {
    'TripPurpose': database.generate_segmentation(
        variable=q02_TripPurpose,
        mapping={
            1: 'business',
            2: 'leisure',
            3: 'attending conf.',
            4: 'business & leisure',
            0: 'unknown',
        },
    ),
    'Gender': database.generate_segmentation(
        variable=q17_Gender,
        mapping={1: 'male', 2: 'female', -1: 'unknown'},
    ),
    'Education': database.generate_segmentation(
        variable=education,
        mapping={
            1: 'high_edu',
            2: 'medium_edu',
            3: 'low_edu',
            -1: 'unknown',
        },
    ),
    'Importance': database.generate_segmentation(
        variable=q11_DepartureOrArrivalIsImportant,
        mapping={1: 'departure', 2: 'arrival', 0: 'not important'},
    ),
    'Who pays': database.generate_segmentation(
        variable=q03_WhoPays,
        mapping={1: 'traveler', 2: 'employer', 3: 'third party', 0: 'unknown'},
    ),
}

tuple_of_segmentations = tuple(all_segmentations.values())


cte_same = Beta('cte_same', 0, None, None, 0)
cte_multiple = Beta('cte_multiple', 0, None, None, 0)
cte_same_catalog, cte_multiple_catalog = segmentation_catalogs(
    generic_name='cte',
    beta_parameters=[cte_same, cte_multiple],
    potential_segmentations=tuple_of_segmentations,
    maximum_number=3,
)

beta_fare = Beta('beta_fare', 0, None, 0, 0)
# The function returns a list. In this case, the list has only one
# element. The variable is followed by a comma so that the only
# catalog in the list is directly extracted and assigned to the
# variable.
(beta_fare_catalog,) = segmentation_catalogs(
    generic_name='beta_fare',
    beta_parameters=[beta_fare],
    potential_segmentations=tuple_of_segmentations,
    maximum_number=3,
)

beta_time = Beta('beta_time', 0, None, 0, 0)

(beta_time_catalog,) = generic_alt_specific_catalogs(
    generic_name='beta_time',
    beta_parameters=[beta_time],
    alternatives=['direct', 'same', 'multiple'],
    potential_segmentations=tuple_of_segmentations,
    maximum_number=3,
)

beta_early = Beta('beta_early', 0, None, 0, 0)
# The function returns a list. In this case, the list has only one
# element. The variable is followed by a comma so that the only
# catalog in the list is directly extracted and assigned to the
# variable.
(beta_early_catalog,) = segmentation_catalogs(
    generic_name='beta_early',
    beta_parameters=[beta_early],
    potential_segmentations=tuple_of_segmentations,
    maximum_number=3,
)

beta_late = Beta('beta_late', 0, None, 0, 0)
# The function returns a list. In this case, the list has only one
# element. The variable is followed by a comma so that the only
# catalog in the list is directly extracted and assigned to the
# variable.
(beta_late_catalog,) = segmentation_catalogs(
    generic_name='beta_late',
    beta_parameters=[beta_late],
    potential_segmentations=tuple_of_segmentations,
    maximum_number=3,
)

beta_legroom = Beta('beta_legroom', 0, 0, None, 0)
# The function returns a list. In this case, the list has only one
# element. The variable is followed by a comma so that the only
# catalog in the list is directly extracted and assigned to the
# variable.
(beta_legroom_catalog,) = segmentation_catalogs(
    generic_name='beta_legroom',
    beta_parameters=[beta_legroom],
    potential_segmentations=tuple_of_segmentations,
    maximum_number=3,
)


utility_direct = (
    beta_fare_catalog * fare_direct_catalog
    + beta_legroom_catalog * legroom_direct_catalog
    + beta_early_catalog * early_direct_catalog
    + beta_late_catalog * late_direct_catalog
    + beta_time_catalog['direct'] * time_direct_catalog
)

utility_same = (
    cte_same_catalog
    + beta_fare_catalog * fare_same_catalog
    + beta_legroom_catalog * legroom_same_catalog
    + beta_early_catalog * early_same_catalog
    + beta_late_catalog * late_same_catalog
    + beta_time_catalog['same'] * time_same_catalog
)

utility_multiple = (
    cte_multiple_catalog
    + beta_fare_catalog * fare_multiple_catalog
    + beta_legroom_catalog * legroom_multiple_catalog
    + beta_early_catalog * early_multiple_catalog
    + beta_late_catalog * late_multiple_catalog
    + beta_time_catalog['multiple'] * time_multiple_catalog
)

V = {
    1: utility_direct,
    2: utility_same,
    3: utility_multiple,
}

# Step 9: availabilities
av = {1: 1, 2: 1, 3: 1}

# Nests
onestop = Beta('mu_onestop', 1, 1, 10, 0), [2, 3]
nonstop = 1.0, [1]
nests_1 = nonstop, onestop

same = Beta('mu_same', 1, 1, 10, 0), [1, 2]
multiple = 1.0, [3]
nests_2 = same, multiple

mu_onestop = Beta('mu_onestop', 1, 1, 10, 0)
mu_same = Beta('mu_same', 1, 1, 10, 0)
alpha_onestop = {1: 1.0, 2: 0.5, 3: 1}
alpha_same = {1: 1.0, 2: 0.5, 3: 1}
nest_onestop = mu_onestop, alpha_onestop
nest_same = mu_same, alpha_same
cnl_nests_1 = nest_onestop, nest_same

alpha = Beta('alpha', 0.5, 0, 1, 0)
mu_onestop = Beta('mu_onestop', 1, 1, 10, 0)
mu_same = Beta('mu_same', 1, 1, 10, 0)
alpha_onestop = {1: 1.0, 2: alpha, 3: 1}
alpha_same = {1: 1.0, 2: 1 - alpha, 3: 1}
nest_onestop = mu_onestop, alpha_onestop
nest_same = mu_same, alpha_same
cnl_nests_2 = nest_onestop, nest_same


model_catalog = Catalog.from_dict(
    catalog_name='model',
    dict_of_expressions={
        'logit': models.loglogit(V, av, chosenAlternative),
        'nested_onestop': models.lognested(V, av, nests_1, chosenAlternative),
        'nested_same': models.lognested(V, av, nests_2, chosenAlternative),
        'CNL_alpha_fixed': models.logcnl_avail(V, av, cnl_nests_1, chosenAlternative),
        'CNL_alpha_est': models.logcnl_avail(V, av, cnl_nests_2, chosenAlternative),
    },
)

the_biogeme = bio.BIOGEME(database, model_catalog)
the_biogeme.modelName = 'airline'



"""File b12panel.py

:author: Michel Bierlaire, EPFL
:date: Sun Apr  9 18:12:17 2023

 Example of a mixture of logit models, using Monte-Carlo integration.
 The datafile is organized as panel data.
"""

import numpy as np
import biogeme.logging as blog
import biogeme.biogeme as bio
from biogeme import models
from biogeme.expressions import (
    Beta,
    bioDraws,
    PanelLikelihoodTrajectory,
    MonteCarlo,
    log,
)
from swissmetro_panel import (
    database,
    CHOICE,
    SM_AV,
    CAR_AV_SP,
    TRAIN_AV_SP,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    SM_TT_SCALED,
    SM_COST_SCALED,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
)

logger = blog.get_screen_logger(level=blog.INFO)
logger.info('Example b12panel.py')

np.random.seed(seed=90267)


# Parameters to be estimated
B_COST = Beta('B_COST', 0, None, None, 0)

# Define a random parameter, normally distributed across individuals,
# designed to be used for Monte-Carlo simulation
B_TIME = Beta('B_TIME', 0, None, None, 0)

# It is advised not to use 0 as starting value for the following parameter.
B_TIME_S = Beta('B_TIME_S', 1, None, None, 0)
B_TIME_RND = B_TIME + B_TIME_S * bioDraws('B_TIME_RND', 'NORMAL_ANTI')

# We do the same for the constants, to address serial correlation.
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_CAR_S = Beta('ASC_CAR_S', 1, None, None, 0)
ASC_CAR_RND = ASC_CAR + ASC_CAR_S * bioDraws('ASC_CAR_RND', 'NORMAL_ANTI')

ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
ASC_TRAIN_S = Beta('ASC_TRAIN_S', 1, None, None, 0)
ASC_TRAIN_RND = ASC_TRAIN + ASC_TRAIN_S * bioDraws('ASC_TRAIN_RND', 'NORMAL_ANTI')

ASC_SM = Beta('ASC_SM', 0, None, None, 1)
ASC_SM_S = Beta('ASC_SM_S', 1, None, None, 0)
ASC_SM_RND = ASC_SM + ASC_SM_S * bioDraws('ASC_SM_RND', 'NORMAL_ANTI')

# Definition of the utility functions
V1 = ASC_TRAIN_RND + B_TIME_RND * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
V2 = ASC_SM_RND + B_TIME_RND * SM_TT_SCALED + B_COST * SM_COST_SCALED
V3 = ASC_CAR_RND + B_TIME_RND * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# Associate utility functions with the numbering of alternatives
V = {1: V1, 2: V2, 3: V3}

# Associate the availability conditions with the alternatives
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# Conditional to the random parameters, the likelihood of one observation is
# given by the logit model (called the kernel)
obsprob = models.logit(V, av, CHOICE)

# Conditional to the random parameters, the likelihood of all observations for
# one individual (the trajectory) is the product of the likelihood of
# each observation.
condprobIndiv = PanelLikelihoodTrajectory(obsprob)

# We integrate over the random parameters using Monte-Carlo
logprob = log(MonteCarlo(condprobIndiv))

# Create the Biogeme object
the_biogeme = bio.BIOGEME(database, logprob)
the_biogeme.modelName = 'b12panel'

# Estimate the parameters.
results = the_biogeme.estimate()
print(results.short_summary())
pandas_results = results.getEstimatedParameters()
print(pandas_results)



"""File b02nonlinear.py

:author: Michel Bierlaire, EPFL
:date: Thu Jul 13 21:31:54 2023

Investigate of nonlinear specifications for the travel time variables:
- linear specification,
- Box-Cox transform,
- power series,
for a total of 3 specifications.
"""
import biogeme.biogeme as bio
from biogeme import models
from biogeme.expressions import Beta
from biogeme.models import boxcox
from biogeme.catalog import Catalog
from results_analysis import report
from swissmetro_data import (
    database,
    CHOICE,
    SM_AV,
    CAR_AV_SP,
    TRAIN_AV_SP,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    SM_TT_SCALED,
    SM_COST_SCALED,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
)


# Parameters to be estimated
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
B_TIME = Beta('B_TIME', 0, None, 0, 0)
B_COST = Beta('B_COST', 0, None, 0, 0)

# Non linear specifications for the travel time

# Parameter of the Box-Cox transform
ell_travel_time = Beta('lambda_travel_time', 1, None, 10, 0)

# Coefficients of the power series
square_tt_coef = Beta('square_tt_coef', 0, None, None, 0)
cube_tt_coef = Beta('cube_tt_coef', 0, None, None, 0)


def power_series(the_variable):
    """Generate the expression of a polynomial of degree 3

    :param the_variable: variable of the polynomial
    :type the_variable: biogeme.expressions.Expression
    """
    return (
        the_variable
        + square_tt_coef * the_variable**2
        + cube_tt_coef * the_variable * the_variable**3
    )


linear_train_tt = TRAIN_TT_SCALED
boxcox_train_tt = boxcox(TRAIN_TT_SCALED, ell_travel_time)
power_train_tt = power_series(TRAIN_TT_SCALED)
train_tt_catalog = Catalog.from_dict(
    catalog_name='train_tt_catalog',
    dict_of_expressions={
        'linear': linear_train_tt,
        'boxcox': boxcox_train_tt,
        'power': power_train_tt,
    },
)

linear_sm_tt = SM_TT_SCALED
boxcox_sm_tt = boxcox(SM_TT_SCALED, ell_travel_time)
power_sm_tt = power_series(SM_TT_SCALED)
sm_tt_catalog = Catalog.from_dict(
    catalog_name='sm_tt_catalog',
    dict_of_expressions={
        'linear': linear_sm_tt,
        'boxcox': boxcox_sm_tt,
        'power': power_sm_tt,
    },
    controlled_by=train_tt_catalog.controlled_by,
)

linear_car_tt = CAR_TT_SCALED
boxcox_car_tt = boxcox(CAR_TT_SCALED, ell_travel_time)
power_car_tt = power_series(CAR_TT_SCALED)
car_tt_catalog = Catalog.from_dict(
    catalog_name='car_tt_catalog',
    dict_of_expressions={
        'linear': linear_car_tt,
        'boxcox': boxcox_car_tt,
        'power': power_car_tt,
    },
    controlled_by=train_tt_catalog.controlled_by,
)


# Definition of the utility functions
V1 = ASC_TRAIN + B_TIME * train_tt_catalog + B_COST * TRAIN_COST_SCALED
V2 = B_TIME * sm_tt_catalog + B_COST * SM_COST_SCALED
V3 = ASC_CAR + B_TIME * car_tt_catalog + B_COST * CAR_CO_SCALED

# Associate utility functions with the numbering of alternatives
V = {1: V1, 2: V2, 3: V3}

# Associate the availability conditions with the alternatives
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# Definition of the model. This is the contribution of each
# observation to the log likelihood function.
logprob = models.loglogit(V, av, CHOICE)

# Create the Biogeme object
the_biogeme = bio.BIOGEME(database, logprob)
the_biogeme.modelName = 'b02nonlinear'
the_biogeme.generate_html = False
the_biogeme.generate_pickle = False

# Estimate the parameters
dict_of_results = the_biogeme.estimate_catalog()

report(dict_of_results)



"""Implements the probability generating function and the CDF of a
cross-nested logit model. This module is not used by Biogeme
itself. It is essentially used for external use. It has been mainly
implemented to verify the analytical derivatives of these functions.

:author: Michel Bierlaire
:date: Fri Apr 22 09:39:49 2022

"""
import numpy as np


def cnl_G(alternatives, nests):
    """Probability generating function and its derivatives

    :param alternatives: a list of alternatives in a given order. In
        principle, the alternative ids should be integers (to be
        consistent with Biogeme), but it may be actually be any object
        for this specific function.
    :type alternatives: list(int)

    :param nests: a tuple containing as many items as nests.
        Each item is also a tuple containing two items:

        - an object of type biogeme.expressions. expr.Expression
          representing the nest parameter,
        - a dictionary mapping the alternative ids with the cross-nested
          parameters for the corresponding nest. If an alternative is
          missing in the dictionary, the corresponding alpha is set to zero.

        Example::

            alphaA = {1: alpha1a,
                      2: alpha2a,
                      3: alpha3a,
                      4: alpha4a,
                      5: alpha5a,
                      6: alpha6a}
            alphaB = {1: alpha1b,
                      2: alpha2b,
                      3: alpha3b,
                      4: alpha4b,
                      5: alpha5b,
                      6: alpha6b}
            nesta = MUA, alphaA
            nestb = MUB, alphaB
            nests = nesta, nestb

    :type nests: tuple

    :return: function that calculates the G function, and its first
        and second derivatives.
    :rtype: f, g, H = fct(np.array(float))

    """
    order = {alt: index for index, alt in enumerate(alternatives)}
    J = len(alternatives)

    def G_and_deriv(y):
        """Probability generating function

        :param y: vector of positive values
        :type y: np.array

        :return: value of the CDF and its derivatives
        :rtype: floap, np.array(float), np.array(np.array(float))
        """

        G = 0.0
        Gi = np.zeros(J)
        Gij = np.zeros((J, J))
        for m in nests:
            mu_m = m[0]
            alphas = m[1]
            nest_specific_sum = 0.0
            for alpha_alt, alpha_value in alphas.items():
                if alpha_value != 0 and y[order[alpha_alt]] != 0:
                    nest_specific_sum += (alpha_value * y[order[alpha_alt]]) ** mu_m
            p1 = (1.0 / mu_m) - 1.0
            p2 = (1.0 / mu_m) - 2.0
            G += nest_specific_sum ** (1.0 / mu_m)
            for i in range(J):
                alpha_i = alphas.get(alternatives[i], 0)
                if alpha_i != 0 and y[i] != 0:
                    Gi[i] += (
                        alpha_i**mu_m * y[i] ** (mu_m - 1) * nest_specific_sum**p1
                    )
                    Gij[i][i] += (1 - mu_m) * nest_specific_sum**p2 * alpha_i ** (
                        2 * mu_m
                    ) * y[i] ** (2 * mu_m - 2.0) + (
                        mu_m - 1
                    ) * nest_specific_sum**p1 * alpha_i**mu_m * y[
                        i
                    ] ** (
                        mu_m - 2
                    )
                    for j in range(i + 1, J):
                        alpha_j = alphas.get(alternatives[j], 0)
                        if alpha_j != 0 and y[j] != 0:
                            Gij[i][j] += (
                                (1 - mu_m)
                                * nest_specific_sum**p2
                                * (alpha_i * alpha_j) ** mu_m
                                * (y[i] * y[j]) ** (mu_m - 1.0)
                            )

        for i in range(J):
            for j in range(i + 1, J):
                Gij[j][i] = Gij[i][j]

        return G, Gi, Gij

    return G_and_deriv


def cnl_CDF(alternatives, nests):
    """Cumulative distribution function and its derivatives

    :param alternatives: a list of alternatives in a given order. In
        principle, the alternative ids should be integers (to be
        consistent with Biogeme), but it may be actually be any object
        for this specific function.
    :type alternatives: list(int)

    :param nests: a tuple containing as many items as nests.
        Each item is also a tuple containing two items:

        - an object of type biogeme.expressions. expr.Expression
          representing the nest parameter,
        - a dictionary mapping the alternative ids with the cross-nested
          parameters for the corresponding nest. If an alternative is
          missing in the dictionaray, the corresponding alpha is set to zero.

        Example::

            alphaA = {1: alpha1a,
                      2: alpha2a,
                      3: alpha3a,
                      4: alpha4a,
                      5: alpha5a,
                      6: alpha6a}
            alphaB = {1: alpha1b,
                      2: alpha2b,
                      3: alpha3b,
                      4: alpha4b,
                      5: alpha5b,
                      6: alpha6b}
            nesta = MUA, alphaA
            nestb = MUB, alphaB
            nests = nesta, nestb

    :type nests: tuple

    :return: function that calculates the CDF, and its first
        and second derivatives.
    :rtype: f, g, H = fct(np.array(float))

    """
    J = len(alternatives)

    G_fct = cnl_G(alternatives, nests)

    def F_and_deriv(xi):
        """Cumulative distribution function

        :param xi: vector of arguments
        :type xi: np.array(float)

        :return: value of the CDF and its derivatives
        :rtype: floap, np.array(float), np.array(np.array(float))
        """
        y = np.where(xi == np.inf, 0, np.exp(-xi))
        G, Gi, Gii = G_fct(y)

        F = np.exp(-G)
        Fi = Gi * y * F
        Fij = np.zeros((J, J))

        for i in range(J):
            Fij[i][i] = F * y[i] * y[i] * (Gi[i] * Gi[i] - Gii[i][i]) - F * Gi[i] * y[i]
            for j in range(i + 1, J):
                Fij[i][j] = F * y[i] * y[j] * (Gi[i] * Gi[j] - Gii[i][j])
        for i in range(J):
            for j in range(i + 1, J):
                Fij[j][i] = Fij[i][j]

        return F, Fi, Fij

    return F_and_deriv



""" Model specification for the choice of restaurants

:author: Michel Bierlaire
:date: Sat Apr 15 16:48:40 2023
"""

from biogeme.expressions import Beta, Variable
from sample import SAMPLE_SIZE as N_ALT

# Parameters to estimate
beta_rating = Beta('beta_rating', 0, None, None, 0)
beta_price = Beta('beta_price', 0, None, None, 0)
beta_chinese = Beta('beta_chinese', 0, None, None, 0)
beta_japanese = Beta('beta_japanese', 0, None, None, 0)
beta_korean = Beta('beta_korean', 0, None, None, 0)
beta_indian = Beta('beta_indian', 0, None, None, 0)
beta_french = Beta('beta_french', 0, None, None, 0)
beta_mexican = Beta('beta_mexican', 0, None, None, 0)
beta_lebanese = Beta('beta_lebanese', 0, None, None, 0)
beta_ethiopian = Beta('beta_ethiopian', 0, None, None, 0)
beta_log_dist = Beta('beta_log_dist', 0, None, None, 0)

# Correction for sampling of alternatives
log_probability = {i: Variable(f'_log_proba_{i}') for i in range(N_ALT)}

# Utility functions. Not that the correction temr is not included in
# the utility functions themselves. It will be included in the call to
# the function loglogit_sampling.

V = {
    i: (
        beta_rating * Variable(f'rating_{i}')
        + beta_price * Variable(f'price_{i}')
        + beta_chinese * Variable(f'category_Chinese_{i}')
        + beta_japanese * Variable(f'category_Japanese_{i}')
        + beta_korean * Variable(f'category_Korean_{i}')
        + beta_indian * Variable(f'category_Indian_{i}')
        + beta_french * Variable(f'category_French_{i}')
        + beta_mexican * Variable(f'category_Mexican_{i}')
        + beta_lebanese * Variable(f'category_Lebanese_{i}')
        + beta_ethiopian * Variable(f'category_Ethiopian_{i}')
        + beta_log_dist * Variable(f'log_dist_{i}')
    )
    for i in range(N_ALT)
}



"""File assisted.py

:author: Michel Bierlaire
:date: Sun Mar 19 17:06:29 2023

New version of the assisted specification using Catalogs

"""
import logging
from biogeme_optimization.vns import vns, ParetoClass
from biogeme_optimization.pareto import Pareto, SetElement, DATE_TIME_STRING
from biogeme_optimization.neighborhood import Neighborhood
from biogeme.biogeme import BIOGEME
from biogeme import tools
from biogeme.parameters import biogeme_parameters
import biogeme.version as bv
import biogeme.exceptions as excep
from biogeme.configuration import Configuration
from biogeme.specification import Specification

logger = logging.getLogger(__name__)


# Operators


class ParetoPostProcessing:
    """Class to process an existing Pareto set."""

    def __init__(
        self,
        biogeme_object,
        pareto_file_name,
    ):
        """Ctor

        :param biogeme_object: object containnig the loglikelihood and the database
        :type biogeme_object: biogeme.biogeme.BIOGEME

        :param pareto_file_name: file where to read and write the Pareto solutions
        :type pareto_file_name: str

        """
        self.biogeme_object = biogeme_object
        self.pareto = Pareto(filename=pareto_file_name)
        self.expression = biogeme_object.loglike
        if self.expression is None:
            error_msg = 'No log likelihood function is defined'
            raise excep.BiogemeError(error_msg)
        self.database = biogeme_object.database
        self.model_names = None

    def reestimate(self, recycle=False):
        """The assisted specification uses quickEstimate to estimate
        the models. A complete estimation is necessary to obtain the
        full estimation results.

        """
        if self.model_names is None:
            self.model_names = tools.ModelNames(prefix=self.biogeme_object.modelName)

        all_results = {}
        for element in self.pareto.pareto:
            config_id = element.element_id
            the_biogeme = BIOGEME.from_configuration(
                config_id=config_id, expression=self.expression, database=self.database
            )
            _ = Configuration.from_string(config_id)
            the_biogeme.modelName = self.model_names(config_id)
            logger.debug(f'REESTIMATE {config_id}')
            the_result = the_biogeme.estimate(recycle=recycle)
            all_results[config_id] = the_result
        return all_results

    def log_statistics(self):
        """Report some statistics about the process in the logger"""
        for msg in self.pareto.statistics():
            logger.info(msg)

    def plot(
        self,
        objective_x=0,
        objective_y=1,
        label_x=None,
        label_y=None,
        margin_x=5,
        margin_y=5,
        ax=None,
    ):
        """Plot the members of the set according to two
            objective functions.  They  determine the x- and
            y-coordinate of the plot.

        :param objective_x: index of the objective function to use for the x-coordinate.
        :type objective_x: int

        :param objective_y: index of the objective function to use for the y-coordinate.
        :type objective_y: int

        :param label_x: label for the x_axis
        :type label_x: str

        :param label_y: label for the y_axis
        :type label_y: str

        :param margin_x: margin for the x axis
        :type margin_x: int

        :param margin_y: margin for the y axis
        :type margin_y: int

        :param ax: matplotlib axis for the plot
        :type ax: matplotlib.Axes

        """

        return self.pareto.plot(
            objective_x, objective_y, label_x, label_y, margin_x, margin_y, ax
        )


class AssistedSpecification(Neighborhood):
    """Class defining assisted specification problem for the VNS algorithm."""

    def __init__(
        self,
        biogeme_object,
        multi_objectives,
        pareto_file_name,
        validity=None,
    ):
        """Ctor

        :param biogeme_object: object containnig the loglikelihood and the database
        :type biogeme_object: biogeme.biogeme.BIOGEME

        :param multi_objectives: function calculating the objectives to minimize
        :type multi_objectives: fct(biogeme.results.bioResults) --> list[float]

        :param pareto_file_name: file where to read and write the Pareto solutions
        :type pareto_file_name: str

        :param validity: function verifying that the estimation results are valid
        :type validity: fct(biogeme.results.bioResults) --> bool

        """
        logger.debug('Ctor assisted specification')
        self.biogeme_object = biogeme_object
        self.central_controller = self.biogeme_object.loglike.set_central_controller()
        Specification.generic_name = biogeme_object.modelName
        self.multi_objectives = staticmethod(multi_objectives)
        self.validity = None if validity is None else staticmethod(validity)
        largest_neighborhood = biogeme_parameters.get_value(
            name='largest_neighborhood', section='AssistedSpecification'
        )
        self.pareto = ParetoClass(
            max_neighborhood=largest_neighborhood, pareto_file=pareto_file_name
        )
        self.pareto.comments = [
            f'Biogeme {bv.getVersion()} [{bv.versionDate}]',
            f'File {self.pareto.filename} created on {DATE_TIME_STRING}',
            f'{bv.AUTHOR}, {bv.DEPARTMENT}, {bv.UNIVERSITY}',
        ]
        self.expression = biogeme_object.loglike
        if self.expression is None:
            error_msg = 'No log likelihood function is defined'
            raise excep.BiogemeError(error_msg)
        self.database = biogeme_object.database
        Specification.expression = self.expression
        Specification.database = self.database
        self.operators = {
            name: self.generate_operator(operator)
            for name, operator in self.central_controller.prepare_operators().items()
        }
        super().__init__(self.operators)

    def generate_operator(self, function):
        """Defines an operator that takes a SetElement as an argument, to
            comply with the interface of the VNS algorithm.

        :param function: one of the function implementing the
            operators from the central controller
        :type function: function(str, int) --> str, int

        :return: operator
        :rtype: function(SetElement, int) --> SetElement, int

        """

        def the_operator(element, step):

            the_new_config, number_of_modifications = function(
                current_config=element.element_id,
                step=step,
            )
            new_specification = Specification(the_new_config)
            return (
                new_specification.get_element(self.multi_objectives),
                number_of_modifications,
            )

        return the_operator

    def is_valid(self, element):
        """Check the validity of the solution.

        :param element: solution to be checked
        :type element: :class:`biogeme.pareto.SetElement`

        :return: valid, why where valid is True if the solution is
            valid, and False otherwise. why contains an explanation why it
            is invalid.
        :rtype: tuple(bool, str)
        """
        if not isinstance(element, SetElement):
            raise excep.BiogemeError(
                f'Wrong type {type(element)} instead of SetElement'
            )

        if self.validity is None:
            return True, None

        specification = Specification.from_string_id(element.element_id)
        results = specification.get_results()
        print(f'{self.validity=}')
        print(f'{results=}')
        return self.validity(results)

    def run(self):
        """Runs the VNS algorithm

        :return: doct with the estimation results of the Pareto optimal models
        :rtype: dict[biogeme.results.bioResults]

        """
        logger.debug('Run assisted specification')
        logger.debug('Pareto solutions BEFORE')
        for elem in self.pareto.pareto:
            logger.debug(elem.element_id)
        # We first try to estimate all possible configurations
        Specification.database = self.biogeme_object.database
        Specification.expression = self.biogeme_object.loglike
        Specification.pareto = self.pareto
        logger.debug('Default specification')
        default_specification = Specification.default_specification()
        print(f'{default_specification=}')
        logger.debug('Default specification: done')
        pareto_before = self.pareto.length_of_all_sets()

        # Check if we can estimate everything
        number_of_specifications = (
            self.biogeme_object.loglike.number_of_multiple_expressions()
        )
        maximum_number = self.biogeme_object.maximum_number_catalog_expressions
        if number_of_specifications <= maximum_number:
            logger.info('We consider all possible combinations of the catalogs.')
            for index, configuration in enumerate(self.biogeme_object.loglike):
                logger.info(f'Model {index}/{number_of_specifications}')
                the_config = configuration.current_configuration()
                the_specification = Specification(the_config)
                the_element = the_specification.get_element(self.multi_objectives)
                Specification.pareto.add(the_element)
                Specification.pareto.dump()
        else:
            logger.info(
                f'The number of possible specifications [{number_of_specifications}] '
                f'exceeds the maximum number [{maximum_number}]. '
                f'A heuristic algorithm is applied.'
            )

            default_element = default_specification.get_element(self.multi_objectives)
            number_of_neighbors = biogeme_parameters.get_value(
                name='number_of_neighbors', section='AssistedSpecification'
            )
            maximum_attempts = biogeme_parameters.get_value(
                name='maximum_attempts', section='AssistedSpecification'
            )
            logger.debug(f'{default_element=}')
            self.pareto = vns(
                problem=self,
                first_solutions=[default_element],
                pareto=self.pareto,
                number_of_neighbors=number_of_neighbors,
                maximum_attempts=maximum_attempts,
            )
        logger.debug('Pareto solutions AFTER')
        for elem in self.pareto.pareto:
            logger.debug(elem.element_id)

        pareto_after = self.pareto.length_of_all_sets()

        self.pareto.dump()
        logger.info(f'Pareto file has been updated: {self.pareto.filename}')
        logger.info(
            f'Before the algorithm: {pareto_before[1]} models, '
            f'with {pareto_before[0]} Pareto.'
        )
        logger.info(
            f'After the algorithm: {pareto_after[1]} models, '
            f'with {pareto_after[0]} Pareto.'
        )

        # Postprocessing
        logger.info(
            'VNS algorithm completed. Postprocessing of the Pareto optimal solutions'
        )

        post_processing = ParetoPostProcessing(
            biogeme_object=self.biogeme_object, pareto_file_name=self.pareto.filename
        )
        estimation_results = post_processing.reestimate()
        post_processing.log_statistics()
        return estimation_results
